import logging
from typing import Optional

import pandas as pd

from mstrio.api import cubes
from mstrio.connection import Connection
from mstrio.modeling import (
    Attribute,
    Expression,
    ExpressionFormat,
    SchemaObjectReference,
)
from mstrio.object_management import SearchPattern, full_search
from mstrio.server import Job
from mstrio.types import ExtendedType, ObjectSubTypes, ObjectTypes
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.helper import (
    exception_handler,
    filter_params_for_func,
    get_valid_project_id,
)
from mstrio.utils.response_processors import cubes as cube_processors
from mstrio.utils.version_helper import (
    class_version_handler,
    is_server_min_version,
    method_version_handler,
)
from mstrio.utils.vldb_mixin import ModelVldbMixin

from .cube import _Cube
from .helpers import (
    AdvancedProperties,
    AttributeTemplateUnit,
    CubeOptions,
    DataPartition,
    Template,
    TimeBasedSettings,
)

logger = logging.getLogger(__name__)


def list_olap_cubes(
    connection: Connection,
    name: str | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list['OlapCube'] | list[dict]:
    """Get list of OlapCube objects or dicts with them. Optionally filter cubes
    by specifying 'name'.
    Optionally use `to_dictionary` to choose output format.
    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name = ?onny will return Sonny and Tonny
    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.
    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.
    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        name (string, optional): value the search pattern is set to, which
            will be applied to the names of olap cubes being searched
        search_pattern (SearchPattern enum or int, optional): pattern to search
            for, such as Begin With or Contains. Possible values are available
            in ENUM mstrio.object_management.SearchPattern.
            Default value is BEGIN WITH (4).
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns OlapCube objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        **filters: Available filter parameters: ['id', 'name', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'owner',
            'ext_type', 'view_media', 'certified_info']
    Returns:
        list with OlapCubes or list of dictionaries
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )

    objects_ = full_search(
        connection,
        object_types=ObjectSubTypes.OLAP_CUBE,
        project=project_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        **filters,
    )
    if to_dictionary:
        return objects_
    return [
        OlapCube.from_dict(source=obj_, connection=connection, with_missing_value=True)
        for obj_ in objects_
    ]


@class_version_handler('11.3.0000')
class OlapCube(ModelVldbMixin, _Cube):
    """Manage single table cube - according to EnumDSSXMLObjectSubTypes its
    subtype is  776 (DssXmlSubTypeReportCube). It inherits all properties from
    Cube.

    Attributes:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Identifier of a pre-existing cube.
        instance_id (str): Identifier of a cube instance if already initialized,
            None by default.
        name (str): Name of the OlapCube.
        size (int): Size of cube.
        status (int): Status of cube.
        path (str): Full path of the cube on environment.
        owner_id (str): ID of cube's owner.
        attributes (list): All attributes of cube.
        metrics (list): All metrics of cube.
        attr_elements (list): All attributes elements of cube.
        selected_attributes (list): IDs of filtered attributes.
        selected_metrics (list): IDs of filtered metrics.
        selected_attr_elements (list): IDs of filtered attribute elements.
        dataframe (DataFrame): Content of a cube extracted into a
            Pandas `DataFrame`.
        template (Template): Template defining OLAP Cube structure with
            references to attributes in rows and to metrics in columns.
        filter (Expression): Filtering element for OLAP Cube, needs to
            be in either 'tree' or 'tokens' format. If both are provided,
            only 'tree' is used.
        options (CubeOptions): Options specifying the behavior of cube.
            Includes possibilities for setting languages, refresh of data
            and partitioning by attribute.
        advanced_properties (AdvancedProperties): Properties containing
            metric join type for every metric and attribute join type for
            every attribute provided in the template.
        time_based_settings (TimeBasedSettings): Settings of timezone and
            calendar objects specified with references by ID.
        show_expression_as (ExpressionFormat): Specify how expressions
            should be presented, default 'ExpressionFormat.TREE'.
            Available values:
            - `ExpressionFormat.TREE`.
                (expression is returned in `text` and `tree` formats)
            - `ExpressionFormat.TOKENS`.
                (expression is returned in `text` and `tokens` formats)
        show_filter_tokens (bool): Specify whether 'filter' is returned in
            'tokens' format, along with `text` and `tree` formats,
            default False.
            - If omitted or False, only `text` and `tree` formats are
                returned.
            - If True, all `text`, 'tree' and `tokens` formats are returned.
    """

    _OBJECT_SUBTYPE = ObjectSubTypes.OLAP_CUBE.value
    _API_GETTERS = {
        **_Cube._API_GETTERS,
        (
            'template',
            'filter',
            'options',
            'advanced_properties',
            'time_based_settings',
        ): None,
    }
    _API_PATCH = {
        **_Cube._API_PATCH,
        (
            'template',
            'filter',
            'options',
            'time_based_settings',
        ): (cube_processors.update, 'partial_put'),
    }
    _PATCH_PATH_TYPES = {
        **_Cube._API_PATCH,
        'template': dict,
        'filter': dict,
        'options': dict,
        'time_based_settings': dict,
    }
    _FROM_DICT_MAP = {
        **_Cube._FROM_DICT_MAP,
        'template': Template.from_dict,
        'filter': Expression.from_dict,
        'options': CubeOptions.from_dict,
        'advanced_properties': AdvancedProperties.from_dict,
        'time_based_settings': TimeBasedSettings.from_dict,
    }
    _MODEL_VLDB_API = {
        'GET_ADVANCED': cubes.get_cube,
        'PUT_ADVANCED': cubes.update_cube,
        'GET_APPLICABLE': cubes.get_applicable_vldb_settings,
    }

    def __init__(
        self,
        connection: Connection,
        id: str,
        name: str | None = None,
        instance_id: str | None = None,
        parallel: bool = True,
        progress_bar: bool = True,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
    ):
        """Initialize an OlapCube instance.

        Note:
            Parameter `name` is not used when fetching. `id` is always used to
            uniquely identify cube.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Identifier of a pre-existing cube.
            name (str): Name of a cube.
            instance_id (str): Identifier of an instance if cube instance has
                already been initialized, None by default.
            parallel (bool, optional): If True (default), utilize optimal number
                of threads to increase the download speed. If False, this
                feature will be disabled.
            progress_bar(bool, optional): If True (default), show the download
                progress bar.
            show_expression_as (ExpressionFormat, str, optional): Specify how
                expressions should be presented,
                default 'ExpressionFormat.TREE'.
                Available values:
                - None.
                    (expression is returned in 'text' format)
                - `ExpressionFormat.TREE` or `tree`.
                    (expression is returned in `text` and `tree` formats)
                - `ExpressionFormat.TOKENS` or `tokens`.
                    (expression is returned in `text` and `tokens` formats)
            show_filter_tokens (bool, optional): Specify whether 'filter'
                is returned in 'tokens' format, along with `text` and `tree`
                formats, default False.
                - If omitted or False, only `text` and `tree` formats are
                    returned.
                - If True, all `text`, 'tree' and `tokens` formats are returned.
        """
        # NOTE use the super.init from _Cube once cube_id is deprecated

        super().__init__(
            connection,
            id=id,
            name=name,
            instance_id=instance_id,
            parallel=parallel,
            progress_bar=progress_bar,
        )

        if is_server_min_version(connection, '11.3.0800'):
            self._set_object_attributes(
                show_expression_as=(
                    show_expression_as
                    if isinstance(show_expression_as, ExpressionFormat)
                    else ExpressionFormat(show_expression_as)
                ),
                show_filter_tokens=show_filter_tokens,
            )

    def _init_variables(self, default_value, **kwargs):
        super()._init_variables(default_value=default_value, **kwargs)
        if is_server_min_version(self.connection, '11.3.0800'):
            self._API_GETTERS[
                (
                    'template',
                    'filter',
                    'options',
                    'advanced_properties',
                    'time_based_settings',
                )
            ] = cube_processors.get
        show_expression_as = kwargs.get('show_expression_as', 'tree')
        self.show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )
        self.show_filter_tokens = kwargs.get('show_filter_tokens', False)
        self.template = (
            Template.from_dict(template)
            if (template := kwargs.get('template'))
            else default_value
        )
        self.filter = (
            Expression.from_dict(filter)
            if (filter := kwargs.get('filter'))
            else default_value
        )
        self.options = (
            CubeOptions.from_dict(options)
            if (options := kwargs.get('options'))
            else default_value
        )
        self.advanced_properties = (
            AdvancedProperties.from_dict(advanced_properties)
            if (advanced_properties := kwargs.get('advanced_properties'))
            else default_value
        )
        self.time_based_settings = (
            TimeBasedSettings.from_dict(time_based_settings)
            if (time_based_settings := kwargs.get('time_based_settings'))
            else default_value
        )
        if template:
            self.__set_definition()

    @classmethod
    def available_metrics(
        cls,
        connection: Connection,
        basic_info_only: bool = True,
        to_dataframe: bool = False,
    ) -> list[dict] | list[pd.DataFrame]:
        """Get all metrics available on I-Server.
        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            basic_info_only(boolean, optional): When True (default value) only
                values of `id`, `name` and `type` will be returned for each
                metric. When False, then all properties of each metric will be
                returned.
            to_dataframe: When False (default value) then metrics are returned
                as a list of dictionaries. When True then metrics are returned
                as Pandas 'DataFrame'.
        Returns:
            List of attributes or attributes as Pandas `DataFrame`.
        """
        return cls.__available_objects(
            connection, ObjectTypes.METRIC, basic_info_only, to_dataframe
        )

    @classmethod
    def available_attributes(
        cls,
        connection: Connection,
        basic_info_only: bool = True,
        to_dataframe: bool = False,
    ) -> list[dict] | list[pd.DataFrame]:
        """Get all attributes available on I-Server.
        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            basic_info_only(boolean, optional): When True (default value) only
                values of `id`, `name` and `type` will be returned for each
                attribute. When False, then all properties of each attribute
                will be returned.
            to_dataframe: When False (default value) then attributes are
                returned as a list of dictionaries. When True then attributes
                are returned as Pandas 'DataFrame'.
        Returns:
            List of attributes or attributes as Pandas `DataFrame`.
        """
        return cls.__available_objects(
            connection, ObjectTypes.ATTRIBUTE, basic_info_only, to_dataframe
        )

    @classmethod
    def available_attribute_forms(
        cls,
        connection: Connection,
        basic_info_only: bool = True,
        to_dataframe: bool = False,
    ) -> list[dict] | list[pd.DataFrame]:
        """Get all attribute forms available on I-Server.
        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            basic_info_only(boolean, optional): When True (default value) only
                values of `id`, `name` and `type` will be returned for each
                attribute form. When False, then all properties of each
                attribute form will be returned.
            to_dataframe: When False (default value) then attribute forms are
                returned as a list of dictionaries. When True then attribute
                forms are returned as Pandas 'DataFrame'.
        Returns:
            List of attribute forms or attribute forms as Pandas `DataFrame`.
        """
        return cls.__available_objects(
            connection, ObjectTypes.ATTRIBUTE_FORM, basic_info_only, to_dataframe
        )

    @classmethod
    def __available_objects(
        cls,
        connection: Connection,
        object_type: ObjectTypes | ObjectSubTypes,
        basic_info_only: bool = True,
        to_dataframe: bool = False,
    ) -> list[dict] | list[pd.DataFrame]:
        """Helper function to get available objects based on their type. It
        should be used to get only available attribute, metrics or attribute
        forms."""
        connection._validate_project_selected()
        avl_objects = full_search(
            connection=connection,
            object_types=object_type,
            project=connection.project_id,
        )
        for a in avl_objects:
            new_type = None
            if a['type'] == ObjectTypes.METRIC.value:
                new_type = 'metric'
            elif a['type'] == ObjectTypes.ATTRIBUTE.value:
                new_type = 'attribute'
            elif a['type'] == ObjectTypes.ATTRIBUTE_FORM.value:
                new_type = 'form'
            if new_type:
                a['type'] = new_type

        if basic_info_only:
            avl_objects = [
                {'id': a['id'], 'name': a['name'], 'type': a['type']}
                for a in avl_objects
            ]
        if to_dataframe:
            avl_objects = pd.DataFrame.from_dict(avl_objects)

        return avl_objects

    @classmethod
    def create(
        cls,
        connection: 'Connection',
        name: str,
        folder_id: str,
        description: str | None = None,
        template: Template | dict | None = None,
        filter: Expression | dict | None = None,
        options: CubeOptions | dict | None = None,
        advanced_properties: AdvancedProperties | dict | None = None,
        time_based_settings: TimeBasedSettings | dict | None = None,
        template_cube: Optional['OlapCube | str'] = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
    ) -> 'OlapCube | None':
        """Create an OLAP Cube by defining its name, destination folder ID,
        description, attributes and metrics or template from 'Template' class to
        specify structure of cube additionally providing filter, options,
        advanced properties and time based settings.

        Note:
            All types of expressions can be represented in the form of tokens,
            but some expressions with type predicate are yet not supported to be
            used for creation or modification in tokens form. For example:
            'expression_nodes.ElementListPredicate' or
            'expression_nodes.FilterQualificationPredicate' and some cases other
            involving embedded objects. Expected errors while be thrown in such
            cases.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name (str): OLAP Cube name.
            folder_id (str): ID of the folder where OLAP Cube should be saved.
            description (str, optional): OLAP Cube description.
            template (Template, dict, optional): Template defining
                OLAP Cube structure with references to attributes in
                rows and to metrics in columns.
            filter (Expression, dict, optional): Filtering element for
                OLAP Cube, needs to be in either 'tree' or 'tokens' format.
                If both are provided, only 'tree' is used.
            options (CubeOptions, dict, optional): Options specifying the
                behavior of cube. Includes possibilities for setting languages,
                refresh of data and partitioning by attribute.
            advanced_properties (AdvancedProperties, dict, optional): Properties
                containing metric join type for every metric and attribute join
                type for every attribute provided in the template.
            time_based_settings (TimeBasedSettings, dict, optional): Settings of
                timezone and calendar objects specified with references by ID.
            template_cube (OlapCube, str, optional): If specified, a newly
                created OLAP Cube will inherit its template from a provided
                OlapCube object. A filter of the referenced cube will be used in
                the case when a 'filter' parameter was not specified.
            show_expression_as (ExpressionFormat, str, optional): Specify how
                expressions should be presented,
                default 'ExpressionFormat.TREE'.
                Available values:
                - None.
                    (expression is returned in 'text' format)
                - `ExpressionFormat.TREE` or `tree`.
                    (expression is returned in `text` and `tree` formats)
                - `ExpressionFormat.TOKENS` or `tokens`.
                    (expression is returned in `text` and `tokens` formats)
            show_filter_tokens (bool, optional): Specify whether 'filter'
                is returned in 'tokens' format, along with `text` and `tree`
                formats, default False.
                - If omitted or False, only `text` and `tree` formats are
                    returned.
                - If True, all `text`, 'tree' and `tokens` formats are returned.

        Raises:
            ValueError: When provided template is empty.

        Returns:
            Newly created OLAP Cube or None in case of incorrectly provided
                arguments.
        """
        return cls.__create(
            connection=connection,
            name=name,
            folder_id=folder_id,
            description=description,
            template=template,
            filter=filter,
            options=options,
            advanced_properties=advanced_properties,
            time_based_settings=time_based_settings,
            template_cube=template_cube,
            show_expression_as=show_expression_as,
            show_filter_tokens=show_filter_tokens,
        )

    @staticmethod
    def __check_objects(objects_: list[dict], obj_name: str) -> bool:
        """Check objects (attribute or metrics) before creation or update of an
        OLAP Cube."""
        return all(OlapCube.__check_object(obj, obj_name) for obj in objects_)

    @staticmethod
    def __check_object(object_: dict, obj_name: str) -> bool:
        """Check a single object (attribute or metric) represented as dictionary
        before creation or update of OLAP Cube. Each dictionary must have keys:
        `id` and `name`. When it has key `type` then it must be the same as
        value of parameter `obj_name`. When this key is missing its value will
        be set to value of parameter `obj_name`."""

        if 'id' not in object_:
            msg = f"{obj_name.capitalize()} '{object_['name']}' is missing key 'id'."
            raise ValueError(msg)

        if 'type' not in object_:
            object_['type'] = obj_name
        elif object_['type'] != obj_name:
            msg = (
                f"Each element in dictionary with {obj_name}s must be of a type "
                f"'{obj_name}'."
            )
            raise ValueError(msg)
        return True

    @staticmethod
    def __check_attributes_update(
        attributes: list[dict], existing_attributes: list[dict], status: int
    ) -> bool:
        """Check dictionaries with attributes before update of an OLAP Cube."""
        return OlapCube.__check_objects_update(
            attributes, existing_attributes, 'attribute', status
        )

    @staticmethod
    def __check_metrics_update(
        metrics: list[dict], existing_metrics: list[dict], status: int
    ) -> bool:
        """Check dictionaries with metrics before update of an OLAP Cube."""
        return OlapCube.__check_objects_update(
            metrics, existing_metrics, 'metric', status
        )

    @staticmethod
    def __check_objects_update(
        objects_: list[dict],
        existing_objects: list[dict],
        object_name: str,
        status: int,
    ) -> bool:
        """Check objects (attributes or metrics) represented as dictionaries
        before update of an OLAP Cube. When status of cube is 0, then it is not
        published and it is possible to freely add or delete objects. Otherwise
        it is possible to only rearrange existing objects. For each object it is
        done also the same check as before creation of an OLAP Cube.
        """
        existing_ids = [o['id'] for o in existing_objects]
        reorganised_objects_count = 0  # to check if all existing objects were provided
        msg = (
            f"It is not possible to add new {object_name}s when editing published cube."
        )
        for object_ in objects_:
            # check if structure of dictionary with an object is correct
            if not OlapCube.__check_object(object_, object_name):
                return False
            # check if status of cube is correct in case of new objects
            if object_['id'] not in existing_ids:
                if status != 0:
                    exception_handler(msg, Warning)
                    return False
            else:
                reorganised_objects_count += 1

        # check if status of cube is correct in case of removing objects
        if reorganised_objects_count != len(existing_ids) and status != 0:
            msg = (
                f"It is not possible delete existing {object_name}s when editing "
                f"published cube."
            )
            exception_handler(msg, Warning)
            return False

        return True

    def publish(self) -> Job:
        """Publish an OLAP Cube. Request to publish an OLAP Cube is an
        asynchronous operation, so the result of it can be seen after calling
        method `refresh_status()` inherited from Cube class.
        Returns:
            Job instance.
        """
        response = cubes.publish(self.connection, self._id)
        logger.info(f"Request for publishing cube '{self.name}' was sent.")
        return Job.from_dict(response.json(), self.connection)

    def export_sql_view(self):
        """Export SQL View of an OLAP Cube.
        Returns:
            SQL View of an OLAP Cube.
        """
        res = cubes.get_sql_view(self._connection, self._id)
        sql_statement = res.json()['sqlStatement']
        return sql_statement

    @classmethod
    @method_version_handler('11.3.0800')
    def __create(
        cls,
        connection: 'Connection',
        name: str,
        folder_id: str,
        template: Template | dict,
        description: str | None,
        filter: Expression | dict | None,
        options: CubeOptions | dict | None,
        advanced_properties: AdvancedProperties | dict | None,
        time_based_settings: TimeBasedSettings | dict | None,
        template_cube: 'OlapCube | str | None',
        show_expression_as: ExpressionFormat | str,
        show_filter_tokens: bool,
    ) -> 'OlapCube | None':
        if not template_cube:
            template = cls.__convert_template_to_dict(template)
        else:
            template_cube = (
                template_cube.id
                if isinstance(template_cube, OlapCube)
                else template_cube
            )

        information = {
            'name': name,
            'description': description,
            'destinationFolderId': folder_id,
        }
        if options:
            options = (
                options.to_dict()
                if isinstance(options, CubeOptions)
                else CubeOptions.from_dict(options).to_dict()
            )
        else:
            options = CubeOptions().to_dict()

        body = {
            'information': information,
            'template': template if not template_cube else None,
            'filter': filter.to_dict() if isinstance(filter, Expression) else filter,
            'options': options,
            'advancedProperties': (
                advanced_properties.to_dict()
                if isinstance(advanced_properties, AdvancedProperties)
                else advanced_properties
            ),
            'timeBased': (
                time_based_settings.to_dict()
                if isinstance(time_based_settings, TimeBasedSettings)
                else time_based_settings
            ),
        }

        data = cube_processors.create(
            connection=connection,
            body=body,
            cube_template_id=template_cube if template_cube else None,
            show_expression_as=get_enum_val(show_expression_as, ExpressionFormat),
            show_filter_tokens=show_filter_tokens,
        )
        data.update(
            show_expression_as=show_expression_as, show_filter_tokens=show_filter_tokens
        )

        logger.info(
            f"Successfully created OLAP Cube named: '{name}' "
            f"with ID: '{data['information']['objectId']}'."
        )

        return cls.from_dict(source=data, connection=connection)

    @method_version_handler('11.3.0800')
    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        abbreviation: str | None = None,
        template: Template | dict | None = None,
        filter: Expression | dict | None = None,
        options: CubeOptions | dict | None = None,
        time_based_settings: TimeBasedSettings | dict | None = None,
    ) -> None:
        """Alter OLAP Cube properties by providing new name, description,
        abbreviation, template, filter, options or time_based_settings
        parameters.

        Note:
            All types of expressions can be represented in the form of tokens,
            but some expressions with type predicate are yet not supported to be
            used for creation or modification in tokens form. For example:
            'expression_nodes.ElementListPredicate' or
            'expression_nodes.FilterQualificationPredicate' and some cases other
            involving embedded objects. Expected errors while be thrown in such
            cases.

        Args:
            name (str, optional): New name for the OLAP Cube.
            description (str, optional): New description for the OLAP Cube.
            abbreviation (str, optional): New abbreviation for the OLAP Cube.
            template (Template, dict, str, optional): Template defining
                OLAP Cube structure with references to attributes in
                rows and to metrics in columns.
            filter (Expression, dict, optional): Filtering element for
                OLAP Cube, needs to be in either 'tree' or 'tokens' format.
                If both are provided, only 'tree' is used.
            options (CubeOptions, dict, optional): Options specifying the
                behavior of cube. Includes possibilities for setting languages,
                refresh of data and partitioning by attribute.
            time_based_settings (TimeBasedSettings, dict, optional): Settings of
                timezone and calendar objects specified with references by ID.

        Raises:
            ValueError: When arguments for altering were not provided or given
                template is empty.
            helper.IServerError: When there was an error on IServer side when
                trying to alter an OLAP Cube.

        Returns:
            None
        """

        if template:
            template = self.__convert_template_to_dict(template)

        arguments = filter_params_for_func(self.alter, locals(), exclude=['self'])

        if not arguments:
            msg = (
                "Please provide at least one of the following parameters to alter: "
                "name, description, abbreviation, template, filter, options or "
                "time_based_settings."
            )
            raise ValueError(msg)

        self._alter_properties(**arguments)

        if template:
            self.__set_definition()

    @method_version_handler('11.3.0800')
    def set_partition_attribute(self, attribute: Attribute | str) -> None:
        """Assign new partition attribute to OLAP Cube.

        Args:
            attribute (Attribute, str): Partition attribute that will
                be set, could be provided by ID or with 'Attribute' object.

        Returns:
            None
        """

        attribute_reference = (
            SchemaObjectReference.create_from(attribute)
            if isinstance(attribute, Attribute)
            else SchemaObjectReference(
                sub_type=ObjectSubTypes.ATTRIBUTE, object_id=attribute
            )
        )

        self.alter(
            options=CubeOptions(
                data_partition=DataPartition(partition_attribute=attribute_reference)
            )
        )

        logger.info(
            "Successfully assigned new partition attribute to "
            f"OLAP Cube named: '{self.name}'."
        )

    @method_version_handler('11.3.0800')
    def remove_partition_attribute(self) -> None:
        """Remove assigned to OLAP Cube partition attribute.

        Returns:
            None
        """

        self.alter(
            options=CubeOptions(data_partition=DataPartition(partition_attribute=None))
        )

        logger.info(
            "Successfully removed assigned partition attribute from "
            f"OLAP Cube named: '{self.name}'."
        )

    @method_version_handler('11.3.0800')
    def list_attribute_forms(self) -> list[dict[list[tuple], str]]:
        """Listing forms for every attribute specified in template. If
        default forms for attribute were used, string indicating that
        will be returned instead of forms.

        Returns:
            List with dicts containing attribute name as key and forms
            as values or string indicating that they are default.
        """

        attributes = [
            elem
            for elem in [
                *self.template.rows,
                *self.template.columns,
                *self.template.page_by,
            ]
            if isinstance(elem, AttributeTemplateUnit)
        ]

        return [
            {
                attribute.name: (
                    'Default forms are used.'
                    if not attribute.forms
                    else [(form.id, form.name) for form in attribute.forms]
                )
            }
            for attribute in attributes
        ]

    @classmethod
    def __convert_template_to_dict(cls, template: Template | dict) -> dict:
        template = template.to_dict() if isinstance(template, Template) else template

        if not (
            template.get('rows') or template.get('columns') or template.get('pageBy')
        ):
            msg = (
                "Please provide not empty template with filled elements in one of "
                "'rows', 'columns' or 'page_by' parameters."
            )
            raise ValueError(msg)

        return template

    def __set_definition(self) -> None:
        definition = self.template.to_dict()
        all_elements = (
            definition.get('rows', [])
            + definition.get('columns', [])
            + definition.get('pageBy', [])
        )
        self._attributes = [
            elem for elem in all_elements if elem['type'] == 'attribute'
        ]
        metrics = [
            elem['elements'] for elem in all_elements if elem['type'] == 'metrics'
        ]
        self._metrics = [] if not metrics else metrics[0]
        self._attr_elements = None
        self.__filter = None
        self.instance_id = None

    def _prepare_instance_body(self) -> dict:
        body = self._instance_config._request_body()

        if self.ext_type in [ExtendedType.RELATIONAL]:
            attributes = {
                elem.id: elem.forms
                for elem in [
                    *self.template.rows,
                    *self.template.columns,
                    *self.template.page_by,
                ]
                if isinstance(elem, AttributeTemplateUnit)
            }
            for attr in body['requestedObjects']['attributes']:
                attr_forms = attributes[attr['id']]
                if isinstance(attr_forms, list):
                    for form in attr_forms:
                        attr['forms'].append(
                            {
                                'id': form.id,
                                'name': form.name,
                            }
                        )
        return body



import logging
from enum import Enum
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import security_filters
from mstrio.modeling import ExpressionFormat
from mstrio.modeling.schema import ObjectSubType, SchemaObjectReference
from mstrio.object_management.folder import Folder
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User, UserGroup
from mstrio.utils.entity import (
    CopyMixin,
    DeleteMixin,
    Entity,
    MoveMixin,
    ObjectSubTypes,
)
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.helper import (
    delete_none_values,
    exception_handler,
    fetch_objects,
    filter_params_for_func,
    find_object_with_name,
    get_valid_project_id,
    get_valid_project_name,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import class_version_handler, method_version_handler

from mstrio.modeling.expression import Expression  # isort:skip

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0200')
def list_security_filters(
    connection: "Connection",
    name_contains: str | None = None,
    to_dictionary: bool = False,
    offset: int | None = None,
    limit: int | None = None,
    project_id: str | None = None,
    project_name: str | None = None,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    show_filter_tokens: bool = False,
    user: User | str | None = None,
    user_group: UserGroup | str | None = None,
    **filters,
) -> list[type["SecurityFilter"]] | list[dict]:
    """Get a list of Security Filter objects or dicts. Optionally filter the
    objects by specifying filters parameter.
    It can also be filtered by user or user group.

    Note: It is not possible to provide both `user` and `user_group` parameter.
        When both arguments are provided error is raised.

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.
    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        name_contains (str, optional): Text that security filter's name
            must contain
        to_dictionary (bool, optional): If `True` returns dictionaries, by
            default (`False`) returns `SecurityFilter` objects.
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior.
        limit (int, optional): Limit the number of elements returned. If `None`
            (default), all objects are returned.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        show_expression_as (ExpressionFormat or str, optional): Specify how
            expressions should be presented
            Available values:
            - None
            (expression is returned in "text" format)
            - `ExpressionFormat.TREE` or `tree`
            (expression is returned in `text` and `tree` formats)
            - `ExpressionFormat.TOKENS` or `tokens`
            (expression is returned in `text` and `tokens` formats)
        show_filter_tokens (bool, optional): Specify whether "qualification"
            is returned in "tokens" format,
            along with `text` and `tree` formats.
            - If omitted or false, only `text` and `tree`
            formats are returned.
            - If true, all `text`, `tree` and `tokens` formats are returned.
        user (str or object, optional): Id of user or `User` object used to
            filter security filters
        user_group (str or object, optional): Id of user group or `UserGroup`
            object used to filter security filters
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'total_users', 'total_user_groups']
    Returns:
        list of security filter objects or list of security filter dictionaries.
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )
    project_name = get_valid_project_name(
        connection=connection,
        project_id=project_id,
    )
    if user and user_group:
        exception_handler(
            msg="You cannot filter by both `user` and `user_group` at the same time.",
            exception_type=ValueError,
        )

    if user:
        user = User(connection, id=user) if isinstance(user, str) else user
        # Filter security filters by user for project defined by
        # `project_name` - assigned based on valid `project_id`
        objects = user.list_security_filters(project_id, to_dictionary=True).get(
            project_name, []
        )
    elif user_group:
        user_group = (
            UserGroup(connection, id=user_group)
            if isinstance(user_group, str)
            else user_group
        )
        # filter security filters by the user group for project defined by
        # `project_name` - assigned based on valid `project_id`
        objects = user_group.list_security_filters(project_id, to_dictionary=True).get(
            project_name, []
        )
    else:
        objects = fetch_objects(
            connection=connection,
            project_id=project_id,
            api=security_filters.get_security_filters,
            limit=limit,
            offset=offset,
            name_contains=name_contains,
            dict_unpack_value='securityFilters',
            filters=filters,
        )
    if to_dictionary:
        return objects

    return [
        SecurityFilter.from_dict(
            {
                "show_expression_as": (
                    show_expression_as
                    if isinstance(show_expression_as, ExpressionFormat)
                    else ExpressionFormat(show_expression_as)
                ),
                "show_filter_tokens": show_filter_tokens,
                **obj,
            },
            connection,
        )
        for obj in objects
    ]


class UpdateOperator(Enum):
    APPLY = "addElements"
    REVOKE = "removeElements"


@class_version_handler('11.3.0100')
class SecurityFilter(Entity, CopyMixin, DeleteMixin, MoveMixin, TranslationMixin):
    """Python representation of MicroStrategy Security Filter object.

    Attributes:
        name: name of the security filter
        id: security filter ID
        description: description of the security filter
        sub_type: string literal used to identify the type of a metadata object,
            ObjectSubType enum
        version: object version ID
        ancestors: list of ancestor folders
        type: object type, ObjectTypes enum
        ext_type: object extended type, ExtendedType enum
        date_created: creation time, DateTime object
        date_modified: last modification time, DateTime object
        owner: User object that is the owner
        acg: access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: object access control list
        is_embedded: if true indicates that the target object of this reference
            is embedded within this object, if this field is omitted
            (as is usual) then the target is not embedded
        path: the path of the object, read only
        primary_locale: the primary locale of the object, in the IETF BCP 47
            language tag format, such as "en-US", read only
        qualification: the security filter definition written as an expression
            tree over predicate nodes, Security Filter support all kinds of
            predicates but bandings
        destination_folder_id: a globally unique identifier used to distinguish
            between metadata objects within the same project
        top_level(list of SchemaObjectReference or list of dicts, optional):
            the top level attribute list
        bottom_level(list of SchemaObjectReference or list of dicts, optional):
            the bottom level attribute list
        hidden: Specifies whether the object is hidden
    """

    _OBJECT_TYPE = ObjectTypes.SECURITY_FILTER
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'sub_type': ObjectSubType,
        'qualification': Expression.from_dict,
        'top_level': [SchemaObjectReference.from_dict],
        'bottom_level': [SchemaObjectReference.from_dict],
    }
    _API_GETTERS = {
        (
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'ancestors',
            'acg',
            'acl',
            'hidden',
            'comments',
        ): objects_processors.get_info,
        (
            'id',
            'name',
            'description',
            'sub_type',
            'date_created',
            'date_modified',
            'path',
            'version_id',
            'is_embedded',
            'primary_locale',
            'qualification',
            'destination_folder_id',
            'top_level',
            'bottom_level',
        ): security_filters.get_security_filter,
        ('users',): security_filters.get_security_filter_members,
    }
    _API_PATCH: dict = {
        (
            'name',
            'description',
            'qualification',
            'destination_folder_id',
            'is_embedded',
            'top_level',
            'bottom_level',
        ): (security_filters.update_security_filter, "put"),
        (
            'folder_id',
            'hidden',
            'comments',
        ): (objects_processors.update, 'partial_put'),
    }
    _PATCH_PATH_TYPES = {
        'name': str,
        'description': str,
        'destination_folder_id': str,
        'qualification': dict,
        'is_embedded': bool,
        'top_level': list,
        'below_level': list,
        'hidden': bool,
    }
    _ALLOW_NONE_ATTRIBUTES = ['qualification']

    def __init__(
        self,
        connection: "Connection",
        id: str | None = None,
        name: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
    ):
        """Initialize security filter object by its identifier.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
            id (str): identifier of a pre-existing security filter containing
                the required data. Defaults to None.
            name (str): name of a pre-existing security filter containing
                required data. Defaults to None.
            show_expression_as (enum or str, optional): specify how expressions
                should be presented
                Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS` or `tokens`
            show_filter_tokens (bool, optional): specify whether `qualification`
                is returned in `tokens` format, along with `text` and `tree`
                format
        """
        connection._validate_project_selected()
        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'name' or 'id' parameter in the constructor."
                )

            security_filter = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_security_filters,
            )
            id = security_filter['id']
        super().__init__(
            connection=connection,
            object_id=id,
            show_expression_as=show_expression_as,
            show_filter_tokens=show_filter_tokens,
        )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.primary_locale = kwargs.get('primary_locale')
        self.is_embedded = kwargs.get('is_embedded')
        self.destination_folder_id = kwargs.get('destination_folder_id')
        self.qualification = (
            Expression.from_dict(kwargs.get('qualification'), self.connection)
            if kwargs.get('qualification')
            else None
        )
        self._sub_type = (
            ObjectSubType(kwargs.get('sub_type')) if kwargs.get('sub_type') else None
        )
        self._path = kwargs.get('path')
        show_expression_as = kwargs.get('show_expression_as', 'tree')
        self.show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )
        self.show_filter_tokens = kwargs.get('show_filter_tokens', False)
        top_level = kwargs.get('top_level')
        self.top_level = (
            [
                SchemaObjectReference.from_dict(level, self.connection)
                for level in top_level
            ]
            if top_level
            else None
        )
        bottom_level = kwargs.get('bottom_level')
        self.bottom_level = (
            [
                SchemaObjectReference.from_dict(level, self.connection)
                for level in bottom_level
            ]
            if bottom_level
            else None
        )
        self.users = None
        self._members = kwargs.get("members")

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        destination_folder: Folder | str,
        qualification: Expression | dict,
        description: str | None = None,
        is_embedded: bool = False,
        primary_locale: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
        top_level: list[dict] | list[SchemaObjectReference] | None = None,
        bottom_level: list[dict] | list[SchemaObjectReference] | None = None,
    ) -> "SecurityFilter":
        """Create a new security filter in a specific project.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            name (str): name of a new security filter
            destination_folder (str or object): a globally unique identifier or
                unique folder name used to distinguish between metadata objects
                within the same project
            qualification (Expression or dict): new security filter
                definition written as an expression tree over predicate nodes.
                It can be provided as `Qualification` object or dictionary.
            description (str, optional): optional description of a new
                security filter
            is_embedded (bool, optional): if true indicates that the target
                object of this reference is embedded within this object
            primary_locale (str, optional): the primary locale of the object,
                in the IETF BCP 47 language tag format, such as "en-US"
            show_expression_as (ExpressionFormat or str, optional): specify how
                expressions should be presented
                Available values:
                - None
                (expression is returned in "text" format)
                - `ExpressionFormat.TREE` or `tree`
                (expression is returned in `text` and `tree` formats)
                - `ExpressionFormat.TOKENS` or `tokens`
                (expression is returned in `text` and `tokens` formats)
            show_filter_tokens (bool, optional): specify whether "qualification"
                is returned in "tokens" format,
                along with `text` and `tree` formats.
                - If omitted or false, only `text` and `tree`
                formats are returned.
                - If true, all `text`, "tree" and `tokens` formats are returned.
            top_level(list of SchemaObjectReference or list of dicts, optional):
                the top level attribute list
            bottom_level(list of SchemaObjectReference or list of dicts,
                optional): the bottom level attribute list

        Returns:
            Security Filter object
        """
        body = {
            "information": {
                "name": name,
                "description": description,
                "destinationFolderId": (
                    destination_folder.id
                    if isinstance(destination_folder, Folder)
                    else destination_folder
                ),
                "primaryLocale": primary_locale,
                "isEmbedded": is_embedded,
            },
            "topLevel": (
                [SchemaObjectReference.to_dict(level) for level in top_level]
                if top_level
                and all(isinstance(level, SchemaObjectReference) for level in top_level)
                else top_level
            ),
            "bottomLevel": (
                [SchemaObjectReference.to_dict(level) for level in bottom_level]
                if bottom_level
                and all(
                    isinstance(level, SchemaObjectReference) for level in bottom_level
                )
                else bottom_level
            ),
        }
        body = delete_none_values(body, recursion=True)
        body["qualification"] = (
            qualification.to_dict()
            if isinstance(qualification, Expression)
            else qualification
        )
        response = security_filters.create_security_filter(
            connection=connection,
            body=body,
            show_expression_as=get_enum_val(show_expression_as, ExpressionFormat),
            show_filter_tokens=show_filter_tokens,
        ).json()
        if config.verbose:
            logger.info(
                f"Successfully created security filter named: '{name}' with ID: '"
                f"{response['id']}'"
            )

        return cls.from_dict(
            source={
                **response,
                'show_expression_as': show_expression_as,
                'show_filter_tokens': show_filter_tokens,
            },
            connection=connection,
        )

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        destination_folder_id: str | None = None,
        qualification: Expression | dict | None = None,
        is_embedded: bool | None = None,
        top_level: list[dict] | list[SchemaObjectReference] | None = None,
        bottom_level: list[dict] | list[SchemaObjectReference] | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
    ):
        """Alter the security filter properties.

        Args:
            name (str, optional): name of a security filter
            description(str, optional): description of a security filter
            destination_folder_id (str, optional): a globally unique identifier
                used to distinguish between objects within the same project
            qualification (Expression, dict, optional): the security filter
            definition written as an expression tree over predicate nodes
            is_embedded (bool, optional): if true indicates that the target
                object of this reference is embedded within this object
            top_level(list of SchemaObjectReference or list of dicts, optional):
                the top level attribute list
            bottom_level(list of SchemaObjectReference or list of dicts,
                optional): the bottom level attribute list
            hidden (bool, optional): Specifies whether the object is hidden.
                Default value: False.
            comments (str, optional): Long description of the security filter.s
        """
        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        self._alter_properties(**properties)

    def apply(
        self,
        users_and_groups: (
            list[User] | list[UserGroup] | list[str] | User | UserGroup | str | None
        ) = None,
    ):
        """Updates members information for a specific security filter.
        Grants a security filter to users or user groups.

        Args:
            users_and_groups (string, object, list of strings or objects): list
            or a single element specifying to which users and user groups
            security filter will be applied. It is possible to provide the whole
            User or UserGroup object(s) or just id(s) of those objects. When
            providing a list it can contain User object(s), UserGroup object(s),
            id(s) at the same time.

        Returns:
            True when applying was successful. False otherwise.
        """
        return self._update_members(UpdateOperator.APPLY, users_and_groups)

    def revoke(
        self,
        users_and_groups: (
            list[User] | list[UserGroup] | list[str] | User | UserGroup | str | None
        ) = None,
    ):
        """Updates members information for a specific security filter.
        Revokes a security filter from users or groups.

        Args:
            users_and_groups (string, object, list of strings or objects): list
            or a single element specifying to which users and user groups
            security filter will be applied. It is possible to provide the whole
            User or UserGroup object(s) or just id(s) of those objects. When
            providing a list it can contain User object(s), UserGroup object(s),
            id(s) at the same time.

        Returns:
            True when applying was successful. False otherwise.
        """
        return self._update_members(UpdateOperator.REVOKE, users_and_groups)

    def _update_members(
        self,
        op: UpdateOperator,
        users_and_groups: (
            list[User] | list[UserGroup] | list[str] | User | UserGroup | str | None
        ) = None,
    ):
        """Update members of security filter."""
        users_or_groups = self._retrieve_ids_from_list(users_and_groups)
        body = {
            "operationList": [
                {"op": op.value, "path": "/members", "value": users_or_groups}
            ]
        }
        res = security_filters.update_security_filter_members(
            self.connection,
            self.id,
            body,
            self.connection.project_id,
            throw_error=False,
        )
        if res.ok:
            self._get_members()
            logger.info(
                f"Successfully updated members for security filter '{self.name}'"
            )
        return res.ok

    @staticmethod
    def _retrieve_ids_from_list(
        objects: (
            list[User] | list[UserGroup] | list[str] | User | UserGroup | str | None
        ) = None,
    ) -> list[str]:
        """Parsing a list which can contain at the same time User object(s),
        UserGroup object(s), id(s) to a list with id(s)."""

        objects = objects if isinstance(objects, list) else [objects]

        return [obj if isinstance(obj, str) else obj.id for obj in objects]

    def _get_members(self):
        """Get the users and user groups that the specified security filter is
        applied to."""
        self._members = []
        self.fetch("users")
        for member in self.users:
            if member['subtype'] == ObjectSubTypes.USER.value:
                self._members.append(User.from_dict(member, self.connection))
            if member['subtype'] == ObjectSubTypes.USER_GROUP.value:
                self._members.append(UserGroup.from_dict(member, self.connection))
        return self._members

    @property
    def sub_type(self):
        return self._sub_type

    @property
    def path(self):
        return self._path

    @property
    def members(self):
        return self._get_members()



import logging
from enum import auto
from pprint import pformat
from typing import Any

from mstrio import config
from mstrio.api import subscriptions
from mstrio.connection import Connection
from mstrio.distribution_services.schedule import Schedule
from mstrio.distribution_services.subscription.content import Content
from mstrio.distribution_services.subscription.delivery import (
    CacheType,
    ClientType,
    Delivery,
    LegacyCacheType,
    LibraryCacheTypes,
    Orientation,
    SendContentAs,
    ShortcutCacheFormat,
    ZipSettings,
)
from mstrio.helpers import NotSupportedError
from mstrio.users_and_groups import User
from mstrio.utils import helper, time_helper
from mstrio.utils.entity import EntityBase
from mstrio.utils.enum_helper import AutoUpperName
from mstrio.utils.helper import (
    get_args_from_func,
    get_default_args_from_func,
    get_valid_project_id,
)
from mstrio.utils.version_helper import method_version_handler

logger = logging.getLogger(__name__)


class RecipientsTypes(AutoUpperName):
    CONTACT_GROUP = auto()
    USER_GROUP = auto()
    CONTACT = auto()
    USER = auto()
    PERSONAL_ADDRESS = auto()
    UNSUPPORTED = auto()


class Subscription(EntityBase):
    """Class representation of MicroStrategy Subscription object.

    Attributes:
        subscription_id: The ID of the Subscription
        connection: The MicroStrategy connection object
        project_id: The ID of the project the Subscription belongs to
    """

    _API_GETTERS = {
        (
            "id",
            "name",
            "multiple_contents",
            "editable",
            "date_created",
            "date_modified",
            "owner",
            "schedules",
            "contents",
            "recipients",
            "delivery",
        ): subscriptions.get_subscription
    }
    _FROM_DICT_MAP = {
        "owner": User.from_dict,
        "contents": (
            lambda source, connection: [
                Content.from_dict(content, connection) for content in source
            ]
        ),
        "delivery": Delivery.from_dict,
        "schedules": (
            lambda source, connection: [
                Schedule.from_dict(content, connection) for content in source
            ]
        ),
        "date_created": time_helper.DatetimeFormats.YMDHMS,
        "date_modified": time_helper.DatetimeFormats.YMDHMS,
    }
    _API_PATCH = [subscriptions.update_subscription]
    _RECIPIENTS_TYPES = RecipientsTypes
    _RECIPIENTS_INCLUDE = ['TO', 'CC', 'BCC', None]

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        subscription_id: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ):
        """Initialize Subscription object, populates it with I-Server data.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted.

        Args:
            connection (Connection): MicroStrategy connection object returned
                by `connection.Connection()`
            id (str, optional): ID of the subscription to be initialized, only
                id or subscription_id have to be provided at once, if both
                are provided id will take precedence
            subscription_id (str, optional): ID of the subscription to be
                initialized
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
        """
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=not project_name,
        )
        if id or subscription_id:
            subscription_id = id if id else subscription_id
        else:
            helper.exception_handler(
                msg='Must specify valid id or subscription_id',
                exception_type=ValueError,
            )

        super().__init__(connection, subscription_id, project_id=project_id)

    def _init_variables(self, project_id, **kwargs):
        super()._init_variables(**kwargs)
        self.subscription_id = kwargs.get('id')
        self.multiple_contents = kwargs.get('multiple_contents')
        self.editable = kwargs.get('editable')
        self.allow_delivery_changes = kwargs.get('allow_delivery_changes')
        self.allow_personalization_changes = kwargs.get('allow_personalization_changes')
        self.allow_unsubscribe = kwargs.get('allow_unsubscribe')
        self.date_created = time_helper.map_str_to_datetime(
            "date_created", kwargs.get("date_created"), self._FROM_DICT_MAP
        )
        self.date_modified = time_helper.map_str_to_datetime(
            "date_modified", kwargs.get("date_modified"), self._FROM_DICT_MAP
        )
        self.owner = (
            User.from_dict(kwargs.get('owner'), self.connection)
            if kwargs.get('owner')
            else None
        )
        self.schedules = (
            [
                Schedule.from_dict(schedule, self._connection)
                for schedule in kwargs.get('schedules')
            ]
            if kwargs.get('schedules')
            else None
        )
        self.contents = (
            [
                Content.from_dict(content, self._connection)
                for content in kwargs.get('contents')
            ]
            if kwargs.get('contents')
            else None
        )
        self.recipients = kwargs.get('recipients')
        self.delivery = (
            Delivery.from_dict(kwargs.get('delivery'))
            if kwargs.get('delivery')
            else None
        )
        self.project_id = project_id

    @method_version_handler('11.3.0000')
    def alter(
        self,  # NOSONAR
        name: str | None = None,
        multiple_contents: bool | None = None,
        allow_delivery_changes: bool | None = None,
        allow_personalization_changes: bool | None = None,
        allow_unsubscribe: bool | None = None,
        send_now: bool = False,
        owner_id: str | None = None,
        schedules: str | list[str] | Schedule | list[Schedule] | None = None,
        contents: Content | None = None,
        recipients: list[str] | list[dict] | None = None,
        delivery: Delivery | dict | None = None,
        delivery_mode: str | None = None,
        custom_msg: str | None = None,
        delivery_expiration_date: str | None = None,
        delivery_expiration_timezone: str | None = None,
        contact_security: bool | None = None,
        filename: str | None = None,
        compress: bool | None = None,
        space_delimiter: str | None = None,
        email_subject: str | None = None,
        email_message: str | None = None,
        email_send_content_as: str | None = None,
        overwrite_older_version: bool | None = None,
        zip_filename: str | None = None,
        zip_password_protect: bool | None = None,
        zip_password: str | None = None,
        file_burst_sub_folder: str | None = None,
        printer_copies: int | None = None,
        printer_range_start: int | None = None,
        printer_range_end: int | None = None,
        printer_collated: bool | None = None,
        printer_orientation: str | None = None,
        printer_use_print_range: bool | None = None,
        cache_cache_type: str | None = None,
        cache_shortcut_cache_format: str | None = None,
        mobile_client_type: str | None = None,
        device_id: str | None = None,
        do_not_create_update_caches: bool | None = None,
        re_run_hl: bool | None = None,
        cache_library_cache_types: list[LibraryCacheTypes | str] = None,
        cache_reuse_dataset_cache: bool = False,
        cache_is_all_library_users: bool = False,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: str | None = None,
    ):
        """
        Alter subscription.

        Args:
            name (str): name of the subscription,
            multiple_contents (bool, optional): whether multiple contents are
                allowed
            allow_delivery_changes (bool): whether the recipients can change
                the delivery of the subscription,
            allow_personalization_changes (bool): whether the recipients can
                personalize the subscription,
            allow_unsubscribe (bool): whether the recipients can unsubscribe
                from the subscription,
            send_now (bool): indicates whether to execute the subscription
                immediately,
            owner_id (str): ID of the subscription owner, by default logged in
                user ID,
            schedules (str | list[str] | Schedule | list[Schedule]):
                Schedules IDs or Schedule objects,
            contents (Content): The content of the subscription.
            recipients (list[str] | list[dict]): list of recipients IDs
                or dicts,
            delivery (Delivery, dict, optional): delivery settings
            delivery_mode (str, enum): the subscription delivery mode [EMAIL,
                FILE, PRINTER, HISTORY_LIST, CACHE, MOBILE, FTP, SNAPSHOT,
                PERSONAL_VIEW, SHARED_LINK, UNSUPPORTED],
            custom_msg (str, optional): customized message displayed when
                Subscription has been successfully altered
            delivery_expiration_date (str): expiration date of the subscription,
                format should be yyyy-MM-dd,
            delivery_expiration_timezone (str, optional): expiration timezone
                of the subscription
            contact_security (bool): whether to use contact security for each
                contact group member,
            filename (str): the filename that will be delivered when
                the subscription is executed,
            compress (bool): whether to compress the file
            space_delimiter (str): space delimiter,
            email_subject (str): email subject associated with the subscription,
            email_message (str): email body of subscription,
            email_send_content_as (str,enum): [data, data_and_history_list,
                data_and_link_and_history_list, link_and_history_list],
            overwrite_older_version (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            zip_filename (str): filename of the compressed content,
            zip_password_protect (bool): whether to password protect zip file,
            zip_password (str): optional password for the compressed file
            file_burst_sub_folder (str): burst sub folder,
            printer_copies (int): the number of copies that should be printed,
            printer_range_start (int): the number indicating the first report
                page that should be printed,
            printer_range_end (int): the number indicating the last report
                page that should be printed,
            printer_collated (bool): whether the printing should be collated,
            printer_orientation (str,enum): [ PORTRAIT, LANDSCAPE ]
            printer_use_print_range (bool): whether print range should be used,
            cache_cache_type (str,enum): [RESERVED, SHORTCUT,
                SHORTCUTWITHBOOKMARK]
            cache_shortcut_cache_format (str,enum): [RESERVED, JSON, BINARY,
                BOTH]
            mobile_client_type (str,enum): [PHONE, TABLET]
            device_id (str): the mobile target project,
            do_not_create_update_caches (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            re_run_hl (bool): whether subscription will re-run against warehouse
            cache_library_cache_types (list[LibraryCacheTypes | str], optional):
                Set of library cache types, available types can be
                web, android, ios
            cache_reuse_dataset_cache (bool): Whether to reuse dataset cache
            cache_is_all_library_users (bool): Whether for all library users
            delivery_notification_enabled (bool): Whether notification is
                enabled, notification applies to cache
            delivery_personal_notification_address_id (str, optional):
                Notification details
        """
        # TODO Potentially remove if new subscription types are supported
        cache_library_cache_types = cache_library_cache_types or [LibraryCacheTypes.WEB]
        if self.delivery.mode in [
            'SNAPSHOT',
            'PERSONAL_VIEW',
            'SHARED_LINK',
            'UNSUPPORTED',
        ]:
            helper.exception_handler(
                msg=f'{self.delivery.mode} subscription altering is not supported.',
                exception_type=NotSupportedError,
            )
        # Schedules logic
        schedules = self.__validate_schedules(schedules=schedules)
        if not schedules:
            schedules = [{'id': sch.id} for sch in self.schedules]

        # Content logic
        if contents:
            contents = self.__validate_contents(contents)
        else:
            contents = [cont.to_dict() for cont in self.contents]

        # Delivery logic
        delivery_expiration_timezone = self.__validate_expiration_time_zone(
            self.connection,
            delivery_expiration_timezone,
        )
        if delivery:
            temp_delivery = (
                Delivery.from_dict(delivery) if isinstance(delivery, dict) else delivery
            )
        else:
            personal_notification_address_id = delivery_personal_notification_address_id

            temp_delivery = self.__change_delivery_properties(
                delivery_mode,
                delivery_expiration_date,
                contact_security,
                email_subject,
                email_message,
                filename,
                compress,
                None,
                zip_filename,
                zip_password,
                zip_password_protect,
                space_delimiter,
                email_send_content_as,
                overwrite_older_version,
                file_burst_sub_folder,
                printer_copies,
                printer_range_start,
                printer_range_end,
                printer_collated,
                printer_orientation,
                printer_use_print_range,
                expiration_time_zone=delivery_expiration_timezone,
                mobile_client_type=mobile_client_type,
                device_id=device_id,
                do_not_create_update_caches=do_not_create_update_caches,
                re_run_hl=re_run_hl,
                cache_type=cache_cache_type,
                shortcut_cache_format=cache_shortcut_cache_format,
                library_cache_types=cache_library_cache_types,
                reuse_dataset_cache=cache_reuse_dataset_cache,
                is_all_library_users=cache_is_all_library_users,
                notification_enabled=delivery_notification_enabled,
                personal_notification_address_id=personal_notification_address_id,
            )
        delivery = temp_delivery.to_dict(camel_case=True)

        # Recipients logic
        recipients = self.__is_val_changed(recipients=recipients)
        recipients = Subscription._validate_recipients(
            self.connection,
            contents,
            recipients,
            self.project_id,
            delivery['mode'],
            self.recipients,
        )

        body = {
            "name": self.__is_val_changed(name=name),
            "allowDeliveryChanges": self.__is_val_changed(
                allow_delivery_changes=allow_delivery_changes
            ),
            "multipleContents": self.__is_val_changed(
                multiple_contents=multiple_contents
            ),
            "allowPersonalizationChanges": self.__is_val_changed(
                allow_personalization_changes=allow_personalization_changes
            ),
            "allowUnsubscribe": self.__is_val_changed(
                allow_unsubscribe=allow_unsubscribe
            ),
            "sendNow": send_now,
            'owner': {
                'id': self.__is_val_changed(nested=self.owner.id, owner_id=owner_id)
            },
            "schedules": schedules,
            "contents": contents,
            "recipients": recipients,
            "delivery": delivery,
        }

        body = helper.delete_none_values(body, recursion=True)

        response = subscriptions.update_subscription(
            self.connection, self.id, self.project_id, body
        )

        if response.ok:
            response = response.json()
            response = helper.camel_to_snake(response)
            self._set_object_attributes(**response)
            if config.verbose:
                msg = f"Updated subscription '{self.name}' with ID: {self.id}."
                msg = custom_msg if custom_msg else msg
                logger.info(msg)

    def __is_val_changed(self, nested=None, **kwargs):
        for key, value in kwargs.items():
            if nested:
                return value if value != nested and value is not None else nested
            else:
                current_val = self.__dict__.get(key)
                # if not current_val: we need to get
                return (
                    value if value != current_val and value is not None else current_val
                )

    @staticmethod
    def __validate_schedules(
        schedules: str | list[str] | Schedule | list[Schedule] = None,
    ):
        tmp_schedules = []
        schedules = schedules if isinstance(schedules, list) else [schedules]
        schedules = [s for s in schedules if s is not None]
        for schedule in schedules:
            if isinstance(schedule, Schedule):
                sch_id = schedule.id
            elif isinstance(schedule, str):
                sch_id = schedule
            tmp_schedules.append({'id': sch_id})

        return tmp_schedules

    @staticmethod
    def __validate_contents(
        contents: list[Content | dict] | Content | dict,
    ) -> list[dict]:
        contents = contents if isinstance(contents, list) else [contents]
        content_type_msg = "Contents must be dictionaries or Content objects."
        return [
            (
                content.to_dict(camel_case=True)
                if isinstance(content, Content)
                else (
                    content
                    if isinstance(content, dict)
                    else helper.exception_handler(content_type_msg, TypeError)
                )
            )
            for content in contents
        ]

    def execute(self):
        """Executes a subscription with given name or GUID for given project."""
        subscriptions.send_subscription(self.connection, self.id, self.project_id)
        if config.verbose:
            logger.info(f"Executed subscription '{self.name}' with ID: {self.id}.")

    @method_version_handler('11.2.0203')
    def delete(self, force: bool = False) -> bool:
        """Delete a subscription. Returns True if deletion was successful.

        Args:
            force: If True, no additional prompt will be shown before deleting
        """
        user_input = 'N'
        if not force:
            user_input = input(
                f"Are you sure you want to delete subscription '{self.name}' with ID: "
                f"{self.id}? [Y/N]: "
            )
        if force or user_input == 'Y':
            response = subscriptions.remove_subscription(
                self.connection, self.id, self.project_id
            )
            if response.ok and config.verbose:
                logger.info(f"Deleted subscription '{self.name}' with ID: {self.id}.")
            return response.ok
        else:
            return False

    @method_version_handler('11.3.0000')
    def available_bursting_attributes(self) -> dict:
        """Get a list of available attributes for bursting feature."""
        contents_bursting = {}
        for content in self.contents:
            response = subscriptions.bursting_attributes(
                self.connection,
                self.project_id,
                content.id,
                content.type.upper(),
            )
            if response.ok:
                contents_bursting[content.id] = response.json()['burstingAttributes']

        return contents_bursting

    @method_version_handler('11.3.0000')
    def available_recipients(self) -> list[dict]:
        """List available recipients for subscription content."""
        body = {"contents": [content.to_dict() for content in self.contents]}
        delivery_type = self.delivery.mode
        response = subscriptions.available_recipients(
            self.connection, self.project_id, body, delivery_type
        )

        if response.ok and config.verbose:
            return response.json()['recipients']

    @method_version_handler('11.3.0000')
    def add_recipient(
        self,
        recipients: list[dict] | dict | list[str] | str = None,
        recipient_id: str | None = None,
        recipient_type: str | None = None,
        recipient_include_type: str = 'TO',
    ):
        """Adds recipient to subscription. You can either specify id, type and
        include_type of single recipient, or just pass recipients list as a
        list of dictionaries.

        Args:
            recipients: list of ids or dicts containing recipients, dict format:
                {"id": recipient_id,
                 "type": "CONTACT_GROUP" / "USER_GROUP" / "CONTACT" /
                         "USER" / "PERSONAL_ADDRESS" / "UNSUPPORTED"
                 "includeType": "TO" / "CC" / "BCC"}
            recipient_id: id of the recipient
            recipient_type: type of the recipient
            recipient_include_type: include type of the recipient one of the
                following: "TO" / "CC" / "BCC", by default "TO"
        """
        if recipients:
            recipients = recipients if isinstance(recipients, list) else [recipients]
        else:
            recipients = []
        if recipient_id and recipient_type:
            recipients.append(
                {
                    "id": recipient_id,
                    "type": recipient_type,
                    "includeType": recipient_include_type,
                }
            )
        elif (recipients == [] and recipient_id is None) or (
            len(recipients) >= 1 and recipient_id
        ):
            msg = (
                "Specify either a recipient ID, type and include type or pass "
                "recipients dictionaries"
            )
            helper.exception_handler(msg, ValueError)

        all_recipients = self.recipients.copy()
        ready_recipients = self.__prepare_recipients(recipients)

        ready_recipients = self._validate_recipients(
            connection=self.connection,
            contents=self.contents,
            recipients=ready_recipients,
            project_id=self.project_id,
            delivery_mode=self.delivery.mode,
            current_recipients=self.recipients,
        )

        if ready_recipients:
            all_recipients.extend(ready_recipients)
            self.alter(recipients=all_recipients)
        elif config.verbose:
            logger.info('No recipients were added to the subscription.')

    @method_version_handler('11.3.0000')
    def remove_recipient(self, recipients: list[str] | list[dict]):
        """Removes recipient from given subscription in given project.

        Args:
            recipients: list of ids or dicts containing recipients, dict format:
                {"id": recipient_id,
                 "type": "CONTACT_GROUP" / "USER_GROUP" / "CONTACT" /
                         "USER" / "PERSONAL_ADDRESS" / "UNSUPPORTED"
                 "includeType": "TO" / "CC" / "BCC"}
        """
        all_recipients = self.recipients
        recipients = recipients if isinstance(recipients, list) else [recipients]
        existing_recipients = [rec['id'] for rec in self.recipients]

        if len(self.recipients) == 1:
            helper.exception_handler(
                "Subscription must have at last one recipient. Add new recipient "
                "before removing.",
                exception_type=ValueError,
            )
        for recipient in recipients:
            rec_id = recipient['id'] if isinstance(recipient, dict) else recipient
            if rec_id not in existing_recipients:
                helper.exception_handler(
                    f"{rec_id} is not a recipient of subscription", UserWarning
                )
            else:
                all_recipients = [rec for rec in all_recipients if rec['id'] != rec_id]
        if len(all_recipients) == 0:
            helper.exception_handler(
                "You cannot remove all existing recipients. Add new recipient before "
                "removing.",
                exception_type=ValueError,
            )
        elif len(self.recipients) - len(all_recipients) > 0:
            self.alter(recipients=all_recipients)
        elif len(self.recipients) == len(all_recipients) and config.verbose:
            logger.info('No recipients were removed from the subscription.')

    def __prepare_recipients(self, recipients):
        existing_recipients = [rec['id'] for rec in self.recipients]
        ready_recipients = []

        def __already_recipient(recipient):
            if recipient in existing_recipients:
                helper.exception_handler(
                    f"{recipient} is already a recipient of subscription", UserWarning
                )
            else:
                ready_recipients.append(recipient)

        for recipient in recipients:
            not_dict_msg = """Wrong recipient format, expected format is
                              {"id": recipient_id,
                               "type": "CONTACT_GROUP" / "USER_GROUP" / "CONTACT" /
                                       "USER" / "PERSONAL_ADDRESS" / "UNSUPPORTED"
                               "includeType": "TO" / "CC" / "BCC" (optional)
                              }"""
            if isinstance(recipient, dict):
                if list(recipient.keys()) in [
                    ['id', 'type', 'includeType'],
                    ['id', 'type'],
                ]:
                    __already_recipient(recipient['id'])
                else:
                    helper.exception_handler(not_dict_msg, TypeError)
            if isinstance(recipient, str):
                __already_recipient(recipient)
        return ready_recipients

    def __change_delivery_properties(  # NOSONAR
        self,  # NOSONAR
        mode: str | None = None,
        expiration: str | None = None,
        contact_security: bool | None = None,
        subject: str | None = None,
        message: str | None = None,
        filename: str | None = None,
        compress: bool | None = None,
        zip_settings: ZipSettings | None = None,
        zip_filename: str | None = None,
        zip_password: str | None = None,
        zip_password_protect: bool | None = None,
        space_delimiter: str | None = None,
        send_content_as: SendContentAs | None = None,
        overwrite_older_version: bool | None = None,
        burst_sub_folder: str | None = None,
        copies: int | None = None,
        range_start: int | None = None,
        range_end: int | None = None,
        collated: bool | None = None,
        orientation: Orientation | None = None,
        use_print_range: bool | None = None,
        expiration_time_zone: str | None = None,
        cache_type: CacheType | None = None,
        shortcut_cache_format: ShortcutCacheFormat | None = None,
        mobile_client_type: ClientType | None = None,
        device_id: str | None = None,
        do_not_create_update_caches: bool | None = None,
        re_run_hl: bool | None = None,
        library_cache_types: list[LibraryCacheTypes | str] = None,
        reuse_dataset_cache: bool = False,
        is_all_library_users: bool = False,
        notification_enabled: bool = False,
        personal_notification_address_id: str | None = None,
    ):
        library_cache_types = library_cache_types or [LibraryCacheTypes.WEB]
        func = self.__change_delivery_properties
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        # create dict of properties to be changed
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]

        # 'zip_settings' is 'zip' in object
        if 'zip_settings' in properties:
            properties['zip'] = properties.pop('zip_settings')

        obj_dict = self.delivery.VALIDATION_DICT
        obj_mode_dict = self.delivery.__dict__[
            self.delivery.mode.lower()
        ].VALIDATION_DICT
        obj_mode_zip_dict = {
            "zip_filename": str,
            "zip_password": str,
            "zip_password_protect": bool,
        }

        if properties:  # at this point only not None values in properties
            for key, value in properties.items():
                if (
                    key in obj_dict
                ):  # Highest level parameters. Mainly mode type and object
                    # Change mode or set attr if it's not mode param
                    if not self.delivery.change_mode(key, value):
                        self.delivery.__setattr__(key, value)
                elif (
                    key in obj_mode_dict
                ):  # parameters of a particular mode e.g. of email
                    helper.rsetattr(
                        self.delivery, f'{self.delivery.mode.lower()}.{key}', value
                    )
                elif key in obj_mode_zip_dict:  # zip settings
                    key = key[4:]
                    if not helper.rgetattr(
                        self.delivery, f'{self.delivery.mode.lower()}.zip', None
                    ):
                        helper.rsetattr(
                            self.delivery,
                            f'{self.delivery.mode.lower()}.zip',
                            ZipSettings(),
                        )
                    helper.rsetattr(
                        self.delivery, f'{self.delivery.mode.lower()}.zip.{key}', value
                    )
        return self.delivery

    @classmethod
    def from_dict(
        cls,
        source: dict[str, Any],
        connection: "Connection" = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> "Subscription":
        """Initialize Subscription object from dictionary.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted"""
        if source.get('project_id') and not project_id:
            project_id = source['project_id']
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
        )
        _source = {
            **source,
            "project_id": project_id,
        }
        obj: Subscription = super().from_dict(_source, connection)

        return obj

    @classmethod
    @method_version_handler('11.3.0000')
    def __create(  # NOSONAR
        cls,  # NOSONAR
        connection: Connection,
        name: str,
        contents: Content | dict,
        project_id: str | None = None,
        project_name: str | None = None,
        multiple_contents: bool | None = None,
        allow_delivery_changes: bool | None = None,
        allow_personalization_changes: bool | None = None,
        allow_unsubscribe: bool | None = None,
        send_now: bool | None = None,
        owner_id: str | None = None,
        schedules: str | list[str] | Schedule | list[Schedule] | None = None,
        recipients: list[dict] | list[str] | None = None,
        delivery: Delivery | dict | None = None,
        delivery_mode: str = Delivery.DeliveryMode.EMAIL,
        delivery_expiration_date: str | None = None,
        delivery_expiration_timezone: str | None = None,
        contact_security: bool = True,
        filename: str | None = None,
        compress: bool = False,
        space_delimiter: str | None = None,
        email_subject: str | None = None,
        email_message: str | None = None,
        email_send_content_as: str = SendContentAs.DATA,
        overwrite_older_version: bool = False,
        zip_filename: str | None = None,
        zip_password_protect: bool | None = None,
        zip_password: str | None = None,
        file_burst_sub_folder: str | None = None,
        printer_copies: int = 1,
        printer_range_start: int = 0,
        printer_range_end: int = 0,
        printer_collated: bool = True,
        printer_orientation: str = Orientation.PORTRAIT,
        printer_use_print_range: bool = False,
        cache_cache_type: CacheType | str = CacheType.RESERVED,
        cache_shortcut_cache_format: (
            ShortcutCacheFormat | str
        ) = ShortcutCacheFormat.RESERVED,
        mobile_client_type: str = ClientType.PHONE,
        device_id: str | None = None,
        do_not_create_update_caches: bool = True,
        re_run_hl: bool = True,
        cache_library_cache_types: list[LibraryCacheTypes | str] = None,
        cache_reuse_dataset_cache: bool = False,
        cache_is_all_library_users: bool = False,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: str | None = None,
    ):
        """Creates a subscription Create_Subscription_Outline.

        Args:
            connection (Connection): a MicroStrategy connection object
            name (str): name of the subscription,
            contents (Content): The content settings.
            project_id (str): project ID,
            project_name (str): project name,
            multiple_contents (bool, optional): whether multiple contents are
                allowed
            allow_delivery_changes (bool): whether the recipients can change
                the delivery of the subscription,
            allow_personalization_changes (bool): whether the recipients can
                personalize the subscription,
            allow_unsubscribe (bool): whether the recipients can unsubscribe
                from the subscription,
            send_now (bool): indicates whether to execute the subscription
                immediately,
            owner_id (str): ID of the subscription owner, by default logged in
                user ID,
            schedules (str | list[str] | Schedule | List[Schedule]):
                Schedules IDs or Schedule objects,
            recipients (list[dict], list[str]): list of recipients IDs or dicts,
            delivery (Delivery | dict): delivery object or dict
            delivery_mode (str, enum): the subscription delivery mode [EMAIL,
                FILE, PRINTER, HISTORY_LIST, CACHE, MOBILE, FTP, SNAPSHOT,
                PERSONAL_VIEW, SHARED_LINK, UNSUPPORTED],
            delivery_expiration_date (str, optional): expiration date of the
                subscription, format should be yyyy-MM-dd,
            delivery_expiration_timezone (str, optional): expiration timezone
                of the subscription, example value 'Europe/London'
            contact_security (bool): whether to use contact security for each
                contact group member,
            filename (str): the filename that will be delivered when
                the subscription is executed,
            compress (bool): whether to compress the file,
            space_delimiter (str): space delimiter,
            email_subject (str): email subject associated with the subscription,
            email_message (str): email body of subscription,
            email_send_content_as (str,enum): [data, data_and_history_list,
                data_and_link_and_history_list, link_and_history_list],
            overwrite_older_version (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            zip_filename (str): filename of the compressed content,
            zip_password_protect (bool): whether to password protect zip file,
            zip_password (str): optional password for the compressed file,
            file_burst_sub_folder (str): burst sub folder,
            printer_copies (int): the number of copies that should be printed,
            printer_range_start (int): the number indicating the first report
                page that should be printed,
            printer_range_end (int): the number indicating the last report
                page that should be printed,
            printer_collated (bool): whether the printing should be collated,
            printer_orientation (str,enum): [ PORTRAIT, LANDSCAPE ]
            printer_use_print_range (bool): whether print range should be used,
            cache_cache_type (str,enum): [RESERVED, SHORTCUT,
                SHORTCUTWITHBOOKMARK]
            cache_shortcut_cache_format (str,enum):
                [RESERVED, JSON, BINARY, BOTH]
            mobile_client_type (str,enum): [PHONE, TABLET]
            device_id (str): the mobile target project,
            do_not_create_update_caches (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            re_run_hl (bool): whether subscription will re-run against warehouse
            cache_library_cache_types (list[LibraryCacheTypes | str], optional):
                Set of library cache types, available types can be
                web, android, ios
            cache_reuse_dataset_cache (bool): Whether to reuse dataset cache
            cache_is_all_library_users (bool): Whether for all library users
            delivery_notification_enabled (bool): Whether notification is
                enabled, notification applies to cache
            delivery_personal_notification_address_id (str, optional):
                Notification details
        """
        cache_library_cache_types = cache_library_cache_types or [LibraryCacheTypes.WEB]
        if connection._iserver_version <= '11.3.0100':
            cache_cache_type = LegacyCacheType[cache_cache_type.name]
        else:
            cache_cache_type = CacheType[cache_cache_type.name]
        name = (
            name
            if len(name) <= 255
            else helper.exception_handler(
                "Name too long. Max name length is 255 characters."
            )
        )
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=not project_name,
        )

        if not schedules:
            msg = "Please specify 'schedules' parameter."
            helper.exception_handler(msg)

        schedules = cls.__validate_schedules(schedules=schedules)

        # Content logic
        contents = contents if isinstance(contents, list) else [contents]
        content_type_msg = "Contents must be dictionaries or Content objects."
        contents = [
            (
                content.to_dict(camel_case=True)
                if isinstance(content, Content)
                else (
                    content
                    if isinstance(content, dict)
                    else helper.exception_handler(content_type_msg, TypeError)
                )
            )
            for content in contents
        ]

        # Delivery logic
        delivery_expiration_timezone = cls.__validate_expiration_time_zone(
            connection,
            delivery_expiration_timezone,
        )
        if delivery:
            temp_delivery = (
                Delivery.from_dict(delivery) if isinstance(delivery, dict) else delivery
            )
        else:
            personal_notification_address_id = delivery_personal_notification_address_id

            temp_delivery = Delivery(
                delivery_mode,
                delivery_expiration_date,
                contact_security,
                email_subject,
                email_message,
                filename,
                compress,
                None,
                zip_password,
                zip_password_protect,
                space_delimiter,
                email_send_content_as,
                overwrite_older_version,
                file_burst_sub_folder,
                printer_copies,
                printer_range_start,
                printer_range_end,
                printer_collated,
                printer_orientation,
                printer_use_print_range,
                mobile_client_type,
                device_id,
                do_not_create_update_caches,
                re_run_hl,
                cache_type=cache_cache_type,
                shortcut_cache_format=cache_shortcut_cache_format,
                library_cache_types=cache_library_cache_types,
                reuse_dataset_cache=cache_reuse_dataset_cache,
                is_all_library_users=cache_is_all_library_users,
                notification_enabled=delivery_notification_enabled,
                personal_notification_address_id=personal_notification_address_id,
                expiration_time_zone=delivery_expiration_timezone,
            )
        delivery = temp_delivery.to_dict(camel_case=True)

        # Recipients logic
        recipients = cls._validate_recipients(
            connection, contents, recipients, project_id, delivery['mode']
        )

        # Create body
        body = {
            "name": name,
            "allowDeliveryChanges": allow_delivery_changes,
            "allowPersonalizationChanges": allow_personalization_changes,
            "allowUnsubscribe": allow_unsubscribe,
            "sendNow": send_now,
            "owner": {"id": owner_id},
            "schedules": schedules,
            "contents": contents,
            "recipients": recipients,
            "delivery": delivery,
        }

        body = helper.delete_none_values(body, recursion=True)
        response = subscriptions.create_subscription(connection, project_id, body)
        if config.verbose:
            unpacked_response = response.json()
            logger.info(
                f"Created subscription '{name}' with ID: '{unpacked_response['id']}'."
            )
        return cls.from_dict(response.json(), connection, project_id)

    @staticmethod
    def _validate_recipients(
        connection: "Connection",
        contents: list[Content | dict],
        recipients: list[str] | list[dict] | str,
        project_id: str,
        delivery_mode: str,
        current_recipients: list[dict] | None = None,
    ):
        def __not_available(recipient):
            if recipient in available_recipients_ids:
                rec = helper.filter_list_of_dicts(available_recipients, id=recipient)
                formatted_recipients.append(rec[0])
            else:
                msg = (
                    f"'{recipient}' is not a valid recipient ID for selected content "
                    "and delivery mode. Available recipients: \n"
                    f"{pformat(available_recipients, indent=2)}"
                )
                helper.exception_handler(msg, ValueError)

        recipients = recipients if isinstance(recipients, list) else [recipients]
        body = {
            "contents": [
                cont.to_dict() if isinstance(cont, Content) else cont
                for cont in contents
            ]
        }
        available_recipients = subscriptions.available_recipients(
            connection, project_id, body, delivery_mode
        )
        if not current_recipients:
            current_recipients = []
        available_recipients = (
            available_recipients.json()['recipients'] + current_recipients
        )
        available_recipients_ids = [rec['id'] for rec in available_recipients]
        # Format recipients list if needed
        formatted_recipients = []
        if recipients:
            for recipient in recipients:
                if isinstance(recipient, dict):
                    __not_available(recipient['id'])
                elif isinstance(recipient, str):
                    __not_available(recipient)
                else:
                    recipient_type = type(recipient)

                    helper.exception_handler(
                        "Recipients must be a dictionaries or a strings, "
                        f"not a {recipient_type}",
                        exception_type=TypeError,
                    )
        return formatted_recipients

    @staticmethod
    def __validate_expiration_time_zone(connection, expiration_time_zone):
        if connection._iserver_version < '11.3.1000' and expiration_time_zone:
            if config.verbose:
                logger.info(
                    'delivery_expiration_timezone argument is available from '
                    'iServer Version 11.3.1000'
                )
        else:
            return expiration_time_zone



import logging
from dataclasses import dataclass

from packaging import version

from mstrio import config
from mstrio.api import subscriptions
from mstrio.connection import Connection
from mstrio.utils.entity import DeleteMixin, EntityBase
from mstrio.utils.helper import (
    Dictable,
    delete_none_values,
    exception_handler,
    fetch_objects_async,
    filter_params_for_func,
    find_object_with_name,
    get_valid_project_id,
)
from mstrio.utils.version_helper import class_version_handler

logger = logging.getLogger(__name__)


def list_dynamic_recipient_lists(
    connection: Connection,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list["DynamicRecipientList"] | list[dict]:
    """Get list of Dynamic Recipient List objects or dicts with them.

    Note:
        When `project_id` is specified, `project_name` is omitted.
        If neither is specified then `project_id` from `connection` object
        is taken.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (str, optional): ID of the project to list the metrics from
        project_name (str, optional): name of the project
        to_dictionary (bool, optional): If True returns list of dictionaries,
            by default (False) returns DynamicRecipientList objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned
        **filters: Available filter parameters: ['name', 'id', 'description',
            'source_report_id', 'physical_address', 'linked_user', 'device']

    Returns:
        list with DynamicRecipientList objects or list of dictionaries
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )

    msg = "Error getting Dynamic Recipient List list."
    chunk_size = (
        1000
        if version.parse(connection.iserver_version) >= version.parse('11.3.0300')
        else 1000000
    )

    objects = fetch_objects_async(
        connection=connection,
        api=subscriptions.list_dynamic_recipient_lists,
        async_api=subscriptions.list_dynamic_recipient_lists_async,
        limit=limit,
        chunk_size=chunk_size,
        filters=filters,
        error_msg=msg,
        dict_unpack_value='listOfDynamicRecipientLists',
        project_id=project_id,
    )

    if to_dictionary:
        return objects
    else:
        return [
            DynamicRecipientList.from_dict(connection=connection, source=obj)
            for obj in objects
        ]


@class_version_handler('11.3.0600')
class DynamicRecipientList(EntityBase, DeleteMixin):
    """Python representation of MicroStrategy DynamicRecipientList object.

    Attributes:
        id: DynamicRecipientList's ID
        name: DynamicRecipientList's name
        description: DynamicRecipientList's description
        source_report_id: Id of the Report that is the source of the
            DynamicRecipientList
        physical_address: Physical Address for the DynamicRecipientList
        linked_user: Linked User for the DynamicRecipientList
        device: Device for the DynamicRecipientList
        recipient_name: Recipient Name for the DynamicRecipientList
        notification_address: Notification Address for the DynamicRecipientList
        notification_device: Notification Device for the DynamicRecipientList
        personalization: Personalization for the DynamicRecipientList
    """

    @dataclass
    class MappingField(Dictable):
        """Python representation of a Mapping Field.

        Attributes:
            attribute_id: ID of the mapped attribute
            attribute_form_id: ID of the mapped attribute's form
        """

        attribute_id: str
        attribute_form_id: str

    _API_GETTERS = {
        (
            'id',
            'name',
            'description',
            'source_report_id',
            'physical_address',
            'linked_user',
            'device',
            'recipient_name',
            'notification_address',
            'notification_device',
            'personalization',
        ): subscriptions.get_dynamic_recipient_list
    }
    _API_PATCH = {
        (
            'id',
            'name',
            'description',
            'source_report_id',
            'physical_address',
            'linked_user',
            'device',
            'recipient_name',
            'notification_address',
            'notification_device',
            'personalization',
        ): (subscriptions.update_dynamic_recipient_list, 'partial_put')
    }
    _API_DELETE = staticmethod(subscriptions.remove_dynamic_recipient_list)
    _FROM_DICT_MAP = {
        **EntityBase._FROM_DICT_MAP,
        'physical_address': MappingField.from_dict,
        'linked_user': MappingField.from_dict,
        'device': MappingField.from_dict,
        'recipient_name': MappingField.from_dict,
        'notification_address': MappingField.from_dict,
        'notification_device': MappingField.from_dict,
        'personalization': MappingField.from_dict,
    }

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> None:
        """Initializes a new instance of a DynamicRecipientList class

        Args:
            connection (Connection): MicroStrategy connection object returned
                by `connection.Connection()`
            id (str, optional): DynamicRecipientList's ID. Defaults to None
            name (str, optional): DynamicRecipientList's name. Defaults to None

        Note:
            Parameter `name` is not used when fetching. If only `name` parameter
            is provided, `id` will be found automatically if such object exists.

        Raises:
            ValueError: if both `id` and `name` are not provided or if
                DynamicRecipientList with the given `name` doesn't exist.
        """
        if not id:
            if name:
                dynamic_recipient_list = find_object_with_name(
                    connection=connection,
                    cls=self.__class__,
                    name=name,
                    listing_function=list_dynamic_recipient_lists,
                )
                id = dynamic_recipient_list['id']
            else:
                exception_handler(
                    msg='Must provide valid id or name', exception_type=ValueError
                )
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=not project_name,
        )
        super().__init__(
            connection=connection, object_id=id, name=name, project_id=project_id
        )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.name = kwargs.get('name')
        self.project_id = kwargs.get('project_id')
        self.description = kwargs.get('description')
        self.source_report_id = kwargs.get('source_report_id')
        self.physical_address = (
            DynamicRecipientList.MappingField.from_dict(paddress)
            if (paddress := kwargs.get('physical_address'))
            else None
        )
        self.linked_user = (
            DynamicRecipientList.MappingField.from_dict(luser)
            if (luser := kwargs.get('linked_user'))
            else None
        )
        self.device = (
            DynamicRecipientList.MappingField.from_dict(dvc)
            if (dvc := kwargs.get('device'))
            else None
        )
        self.recipient_name = (
            DynamicRecipientList.MappingField.from_dict(rname)
            if (rname := kwargs.get('recipient_name'))
            else None
        )
        self.notification_address = (
            DynamicRecipientList.MappingField.from_dict(naddress)
            if (naddress := kwargs.get('notification_address'))
            else None
        )
        self.notification_device = (
            DynamicRecipientList.MappingField.from_dict(ndevice)
            if (ndevice := kwargs.get('notification_device'))
            else None
        )
        self.personalization = (
            DynamicRecipientList.MappingField.from_dict(prsnlz)
            if (prsnlz := kwargs.get('personalization'))
            else None
        )

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        source_report_id: str,
        physical_address: MappingField,
        linked_user: MappingField,
        device: MappingField,
        project_id: str | None = None,
        project_name: str | None = None,
        description: str | None = None,
        recipient_name: MappingField | None = None,
        notification_address: MappingField | None = None,
        notification_device: MappingField | None = None,
        personalization: MappingField | None = None,
    ) -> "DynamicRecipientList":
        """Create a new DynamicRecipientList with specified properties.

        Args:
            name (string): DynamicRecipientList's name
            source_report_id (string): ID of the Report that is the source of
                the DynamicRecipientList
            physical_address (MappingField): Mapping Field representing the
                Physical Address for the DynamicRecipientList
            linked_user (MappingField): Mapping Field representing the Linked
                User for the DynamicRecipientList
            device (MappingField): Mapping Field representing the Device for
                the DynamicRecipientList
            project_id (string, optional): ID of the project in which the
                DynamicRecipientList is to be created, if not provided, ID will
                be taken from project_name, if neither is provided, project will
                be extracted from the connection
            project_name (string, optional): name of the project in which the
                DynamicRecipientList is to be created
            description (string, optional): DynamicRecipientList's description
            recipient_name (MappingField, optional): Mapping Field representing
                the Recipient Name for the DynamicRecipientList
            notification_address (MappingField, optional): Mapping Field
                representing the Notification Address for the
                DynamicRecipientList
            notification_device (MappingField, optional): Mapping Field
                representing the Notification Device for the
                DynamicRecipientList
            personalization (MappingField, optional): Mapping Field representing
                the Personalization for the DynamicRecipientList

        Returns:
            DynamicRecipientList class object.
        """
        body = {
            'name': name,
            'description': description if description else None,
            'sourceReportId': source_report_id,
            'physicalAddress': physical_address.to_dict(),
            'linkedUser': linked_user.to_dict(),
            'device': device.to_dict(),
            'recipientName': recipient_name.to_dict() if recipient_name else None,
            'notificationAddress': (
                notification_address.to_dict() if notification_address else None
            ),
            'notificationDevice': (
                notification_device.to_dict() if notification_device else None
            ),
            'personalization': personalization.to_dict() if personalization else None,
        }
        body = delete_none_values(source=body, recursion=True)
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=not project_name,
        )
        response = subscriptions.create_dynamic_recipient_list(
            connection=connection, project_id=project_id, body=body
        ).json()

        if config.verbose:
            logger.info(
                f"Created Dynamic Recipient List named: '{name}' with ID: '"
                f"{response['id']}'"
            )

        return cls.from_dict(source={**response}, connection=connection)

    def alter(
        self,
        name: str | None = None,
        source_report_id: str | None = None,
        physical_address: MappingField | None = None,
        linked_user: MappingField | None = None,
        device: MappingField | None = None,
        description: str | None = None,
        recipient_name: MappingField | None = None,
        notification_address: MappingField | None = None,
        notification_device: MappingField | None = None,
        personalization: MappingField | None = None,
    ) -> None:
        """Alter a DynamicRecipientList's specified properties

        Note:
            If one alters the source_report_id, all existing Mapping Fields
            also need to be updated to reflect the new source report.

        Args:
            name (string, optional): DynamicRecipientList's name
            source_report_id (string, optional): ID of the Report that is the
                source of the DynamicRecipientList
            physical_address (MappingField, optional): Mapping Field
                representing the Physical Address for the DynamicRecipientList
            linked_user (MappingField, optional): Mapping Field representing
                the Linked User for the DynamicRecipientList
            device (MappingField, optional): Mapping Field representing the
                Device for the DynamicRecipientList
            description (string, optional): DynamicRecipientList's description
            recipient_name (MappingField, optional): Mapping Field representing
                the Recipient Name for the DynamicRecipientList
            notification_address (MappingField, optional): Mapping Field
                representing the Notification Address for the
                DynamicRecipientList
            notification_device (MappingField, optional): Mapping Field
                representing the Notification Device for the
                DynamicRecipientList
            personalization (MappingField, optional): Mapping Field
                representing the Personalization for the DynamicRecipientList
        """
        name = name or self.name
        description = description or self.description
        source_report_id = source_report_id or self.source_report_id
        physical_address = physical_address or self.physical_address
        linked_user = linked_user or self.linked_user
        device = device or self.device
        recipient_name = recipient_name or self.recipient_name
        notification_address = notification_address or self.notification_address
        notification_device = notification_device or self.notification_device
        personalization = personalization or self.personalization
        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        self._alter_properties(**properties)



"""This is the demo script to show how to manage cube caches.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.project_objects import (
    CubeCache, delete_cube_cache, delete_cube_caches, list_cube_caches, OlapCube
)
from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert project name here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Define a variable which can be later used in a script
CUBE_ID = $cube_id  # Insert name of cube on which you want to perform actions

# get caches from an OLAP Cube
oc_caches = OlapCube(conn, id=CUBE_ID).get_caches()

# Define variables which can be later used in a script
NODE_NAME = $node_name  # Insert name of node on which you want to perform actions
DB_CONNECTION_ID = $db_connection_id  # insert ID of DB connection you want to include in your functions

# list all caches
list_cube_caches(connection=conn)
# list all caches on a given node
list_cube_caches(connection=conn, nodes=NODE_NAME)
# list all loaded caches on a given node
list_cube_caches(connection=conn, nodes=NODE_NAME, loaded=True)
# list all cache on a given node for given cube
list_cube_caches(connection=conn, nodes=NODE_NAME, cube_id=CUBE_ID)
# list all cache on a given node for given database connection
list_cube_caches(connection=conn, nodes=NODE_NAME, db_connection_id=DB_CONNECTION_ID)

# Define a variable which can be later used in a script
CUBE_CACHE_ID = $cube_cache_id  # Insert ID of cube cache that you want to perform your actions upon

# get a single cube cache by its id
cube_cache_ = CubeCache(connection=conn, cache_id=CUBE_CACHE_ID)

# unload a cube cache
cube_cache_.unload()
# load a cube cache
cube_cache_.load()
# deactivate a cube cache
cube_cache_.deactivate()
# activate a cube cache
cube_cache_.activate()

# refresh cube cache
cube_cache_.fetch()

# get state of cube cache
cc_state = cube_cache_.state

# get properties of cube cache
cc_properties = cube_cache_.list_properties()

# Delete a single cube cache (there are two ways). When `force` argument is set
# to `False` (default value) then deletion must be approved.
cube_cache_.delete(force=True)
delete_cube_cache(connection=conn, id=cube_cache_.id, force=True)

# delete all cube caches (the same rule with `force` as above)
delete_cube_caches(connection=conn, force=True)
# delete all cube caches on a given node (the same rule with `force` as above)
delete_cube_caches(connection=conn, nodes=NODE_NAME, force=True)
# delete all loaded cube caches on a given node (the same rule with `force` as
# above)
delete_cube_caches(connection=conn, nodes=NODE_NAME, loaded=True, force=True)
# delete all cube caches on a given node for a given cube (the same rule with
# `force` as above)
delete_cube_caches(connection=conn, nodes=NODE_NAME, cube_id=CUBE_ID, force=True)
# delete all cube caches on a given node for a given database connection (the
# same rule with `force` as above)
delete_cube_caches(connection=conn, nodes=NODE_NAME, db_connection_id=DB_CONNECTION_ID, force=True)



from mstrio.connection import Connection
from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler


@unpack_information
@ErrorHandler(err_msg="Error reading fact with ID: {id}.")
def read_fact(
    connection: 'Connection',
    id: str,
    project_id: str = None,
    changeset_id: str = None,
    show_expression_as: str | None = None,
    show_potential_tables: bool = False,
    show_fields: str | None = None,
):
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id
    spt = (
        str(show_potential_tables).lower()
        if show_potential_tables is not None
        else None
    )
    return connection.get(
        endpoint=f'/api/model/facts/{id}',
        headers={
            'X-MSTR-ProjectID': project_id,
            'X-MSTR-MS-Changeset': changeset_id,
        },
        params={
            'showExpressionAs': show_expression_as,
            'showPotentialTables': spt,
            'showFields': show_fields,
        },
    )


@unpack_information
@ErrorHandler(err_msg="Error creating a fact.")
def create_fact(
    connection: 'Connection',
    body: dict,
    show_expression_as: str | None = None,
    show_potential_tables: bool = False,
):
    spt = (
        str(show_potential_tables).lower()
        if show_potential_tables is not None
        else None
    )
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            endpoint='/api/model/facts',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            params={
                'showExpressionAs': show_expression_as,
                'showPotentialTables': spt,
            },
            json=body,
        )


@unpack_information
@ErrorHandler(err_msg="Error updating fact with ID: {id}.")
def update_fact(
    connection: 'Connection',
    id: str,
    body: dict,
    show_expression_as: str | None = None,
    show_potential_tables: bool = False,
):
    spt = (
        str(show_potential_tables).lower()
        if show_potential_tables is not None
        else None
    )
    with changeset_manager(connection) as changeset_id:
        return connection.put(
            endpoint=f'/api/model/facts/{id}',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            params={
                'showExpressionAs': show_expression_as,
                'showPotentialTables': spt,
            },
            json=body,
        )



import logging
from dataclasses import dataclass
from enum import auto
from typing import TYPE_CHECKING, Optional

from mstrio.modeling.schema.helpers import SchemaObjectReference
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from mstrio.connection import Connection


@dataclass
class DimensionalityUnit(Dictable):
    """
    Object that specifies the dimensionality unit.

    Attributes:
        dimtyUnitType: Type of unit in dimensionality
        target: target of dimensionality
        aggregation: type of aggregation in dimensionality
        filtering: Type of filtering applied in this dimensionality
        relativePosition: number showing relative position of dimensionality
        in metric reports
        target: Object reference to attribute or report, on which the
        dimensionality will be grouped at
        axisCollection: dict of numbers showing, at which axis
        metric will be shown
    """

    class DimensionalityUnitType(AutoName):
        DEFAULT = auto()
        ATTRIBUTE = auto()
        DIMENSION = auto()
        REPORT_LEVEL = auto()
        REPORT_BASE_LEVEL = auto()
        ROLE = auto()

    class Aggregation(AutoName):
        NORMAL = auto()
        FIRST_IN_FACT = auto()
        LAST_IN_FACT = auto()
        FIRST_IN_RELATIONSHIP = auto()
        LAST_IN_RELATIONSHIP = auto()

    class Filtering(AutoName):
        APPLY = auto()
        ABSOLUTE = auto()
        IGNORE = auto()
        NONE = auto()

    _FROM_DICT_MAP = {
        'dimensionality_unit_type': DimensionalityUnitType,
        'filtering': Filtering,
        'aggregation': Aggregation,
        'target': SchemaObjectReference.from_dict,
    }

    dimensionality_unit_type: DimensionalityUnitType
    aggregation: Aggregation | None = None
    filtering: Filtering | None = None
    group_by: bool | None = None
    relative_position: int | None = None
    target: SchemaObjectReference | None = None
    axis_collection: dict | None = None

    @classmethod
    def from_dict(cls, source: dict):
        new_source = source.copy()
        new_source['dimensionality_unit_type'] = source.get('dimtyUnitType', [])
        return super().from_dict(new_source)

    def to_dict(cls, camel_case: bool = True) -> dict:
        obj_dict = super().to_dict(camel_case=camel_case)
        if obj_dict.get('dimensionalityUnitType'):
            obj_dict['dimtyUnitType'] = obj_dict.pop('dimensionalityUnitType')
        return obj_dict


@dataclass
class Dimensionality(Dictable):
    """
    Object that specifies the dimensionality.

    This class can only be used for simple metrics.
    For creation and altering of compound metrics, dimensionality needs to be
    defined through the use of the Expression class.

    Attributes:
        dimtyUnits: The list of dimty unit
        excludeAttribute: boolean whether to exclude attribute
            absent in report or level(dimensionality)
        allowAddingUnit: boolean whether to allow other users
            to add extra units
        prompt: reference to a prompt
    """

    _FROM_DICT_MAP = {
        'dimensionality_units': [DimensionalityUnit],
        'prompt': SchemaObjectReference,
    }

    dimensionality_units: list[DimensionalityUnit] | None = None
    exclude_attribute: bool | None = None
    allow_adding_unit: bool | None = None
    prompt: SchemaObjectReference | None = None

    @classmethod
    def from_dict(
        cls,
        source: dict,
        connection: Optional['Connection'] = None,
        to_snake_case: bool = True,
    ):
        data = source.copy()
        data['dimensionality_units'] = data.get('dimty_units', [])
        return super().from_dict(data, connection, to_snake_case)

    def to_dict(cls, camel_case: bool = True) -> dict:
        obj_dict = super().to_dict(camel_case=camel_case)
        if obj_dict.get('dimensionalityUnits'):
            obj_dict['dimtyUnits'] = obj_dict.pop('dimensionalityUnits')
        return obj_dict



import logging

from tqdm import tqdm

from mstrio import config
from mstrio.api import tables as tables_api
from mstrio.connection import Connection
from mstrio.modeling.schema import ObjectSubType, SchemaObjectReference
from mstrio.modeling.schema.attribute import Attribute
from mstrio.modeling.schema.fact import Fact
from mstrio.modeling.schema.helpers import (
    PhysicalTableType,
    TableColumn,
    TableColumnMergeOption,
    TablePrefixOption,
)
from mstrio.modeling.schema.table.physical_table import PhysicalTable
from mstrio.object_management import Folder
from mstrio.object_management.search_operations import full_search
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import get_enum, get_enum_val
from mstrio.utils.helper import (
    delete_none_values,
    exception_handler,
    fetch_objects,
    get_args_from_func,
    get_default_args_from_func,
    get_valid_project_id,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.response_processors import tables
from mstrio.utils.translation_mixin import TranslationMixin
from mstrio.utils.version_helper import class_version_handler, method_version_handler

NO_PROJECT_ERR_MSG = "You must specify or select a project."

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0100')
def list_logical_tables(
    connection: Connection,
    to_dictionary: bool = False,
    table_type: PhysicalTableType | None = None,
    name: str | None = None,
    folder_id: str | None = None,
    folder_name: str | None = None,
    project_id: str | None = None,
    project_name: str | None = None,
    filters: dict | None = None,
    limit: int | None = None,
) -> list["LogicalTable"] | list[dict]:
    """List all logical tables in a project mapped to a specified connection.
       Optionally, you can filter by physical table type.

    Args:
        connection (Connection): Object representation of MSTR Connection.
        name (Optional[str], optional): Name of a table.
        project_id (Optional[str], optional): ID of a project. Defaults to None.
        project_name (Optional[str], optional): Name of a project. Defaults to
            None.
        folder_id (Optional[str], optional): ID of a folder in which to look for
            logical tables. Defaults to None.
        folder_name (Optional[str], optional): Name of a folder in which to look
            for logical tables. Defaults to None.
        to_dictionary (bool, optional): If True returns a list of dictionaries.
            Defaults to False.
        table_type(PhysicalTableType, optional): If specified, returns a list
            of logical tables with physical table with this type.
        filters: dict that specifies filter expressions. Only used if a name,
            folder_id, folder_name, project_id or project_name are provided.
            Available filters are:
            ['date_created', 'date_modified', 'version_id', 'acg',
             'primary_locale', 'sub_type', 'name', 'id']
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.

    Returns:
        list["LogicalTable"] | list[dict]: A list of LogicalTable objects
            or dictionaries representing logical tables.
    """
    if any([name, folder_id, folder_name, project_id, project_name]):
        return _full_search_logical_tables(
            connection=connection,
            to_dictionary=to_dictionary,
            name=name,
            folder_id=folder_id,
            folder_name=folder_name,
            project_id=project_id,
            project_name=project_name,
            limit=limit,
        )
    else:
        if table_type:
            logical_tables = fetch_objects(
                connection,
                tables_api.get_tables,
                limit=None,
                filters=filters or {},
                dict_unpack_value="tables",
                fields="information,physicalTable",
            )
            table_type = get_enum_val(table_type, PhysicalTableType)

            def remove_physical_table(data):
                data.pop('physical_table')
                return data

            logical_tables = [
                remove_physical_table(table)
                for table in logical_tables
                if table.get("physical_table").get("type") == table_type
            ]
            logical_tables = logical_tables[:limit] if limit else logical_tables
        else:
            logical_tables = fetch_objects(
                connection,
                tables_api.get_tables,
                limit=limit,
                filters=filters or {},
                dict_unpack_value="tables",
                fields="information",
            )

        return (
            logical_tables
            if to_dictionary
            else LogicalTable.bulk_from_dict(
                source_list=logical_tables, connection=connection
            )
        )


def _full_search_logical_tables(
    connection: Connection,
    name: str | None = None,
    folder_id: str | None = None,
    folder_name: str | None = None,
    project_id: str | None = None,
    project_name: str | None = None,
    limit: int | None = None,
    to_dictionary: bool = False,
) -> list[type["LogicalTable"]] | list[dict]:
    """Searches for a logical table in a specified project. You can narrow the
       search result by folder's id or name.

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.

    Args:
        connection (Connection): Object representation of MSTR Connection.
        name (Optional[str], optional): Name of a table.
        project_id (Optional[str], optional): ID of a project. Defaults to None.
        project_name (Optional[str], optional): Name of a project. Defaults to
            None.
        folder_id (Optional[str], optional): ID of a folder in which to look for
            logical tables. Defaults to None.
        folder_name (Optional[str], optional): Name of a folder in which to look
            for logical tables. Defaults to None.
        to_dictionary (bool, optional): If True, the function will return list
            of dictionaries. Otherwise, list of LogicalTable objects.
            Defaults to False.

    Throws:
        ValueError if:
            - Project could not be determined
            - `folder_name` was specified but folder could not be found.

    Returns:
        list[dict] | list["LogicalTable"]: A list of logical tables
            in a given project represented either as objects or dictionaries.

    Examples:
        >>> full_search_logical_tables(connection)
        >>> full_search_tables(connection,
                               project_name='MicroStrategy Tutorial',
                               folder_id='24BAE4BF704E8E269333BEAF7195AF88')
        >>> full_search_logical_tables(connection,
                                       project_name='MicroStrategy Tutorial',
                                       folder_name='My Answers')
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )

    if folder_name and not folder_id:
        folders = full_search(
            connection=connection,
            project=project_id,
            object_types=ObjectTypes.FOLDER,
            limit=limit,
        )
        folder = next(
            filter(lambda _folder: _folder.get("name") == folder_name, folders), None
        )
        folder_id = (
            folder["id"]
            if folder
            else exception_handler(
                'Folder with a given name was not found.', exception_type=ValueError
            )
        )

    logical_tables = full_search(
        connection=connection,
        name=name,
        project=project_id,
        object_types=ObjectTypes.TABLE,
        root=folder_id,
        limit=limit,
    )
    return (
        logical_tables
        if to_dictionary
        else LogicalTable.bulk_from_dict(
            source_list=logical_tables, connection=connection
        )
    )


@method_version_handler('11.3.0100')
def list_changeset_tables(
    connection: Connection,
    project_id: str | None = None,
    project_name: str | None = None,
    limit: int | None = None,
    changeset_id: str | None = None,
    **filters,
) -> list[type["LogicalTable"]]:
    # Changeset must be specified
    if not changeset_id:
        exception_handler(
            msg="You must specify changeset ID.", exception_type=ValueError
        )

    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=not project_name,
    )

    tables_list = fetch_objects(
        connection=connection,
        api=tables_api.get_tables,
        limit=limit,
        project_id=project_id,
        changeset_id=changeset_id,
        dict_unpack_value="tables",
        filters=filters,
    )

    return LogicalTable.bulk_from_dict(source_list=tables_list, connection=connection)


@class_version_handler('11.3.0100')
class LogicalTable(Entity, DeleteMixin, MoveMixin, TranslationMixin):
    """An object representation of a logical table, referred to as Table in
    Command Manager. A logical table describes the higher-level data model
    objects (facts, attributes, etc.) that the architect wishes to use to model
    the contents of the physical table.

    A logical table cannot exist without a physical table. Each logical table
    is mapped to exactly one physical table. However, a physical table may be
    mapped to many logical tables (each additional one is called logical table
    alias). This class exposes interface to create both logical and logical
    table alias.

    After performing a write operation on a LogicalTable object, you may want to
    reload the schema using mstrio.modeling.schema.schema_management module.

    Attributes:
        name: name of the logical table
        id: id of the logical table
        description: description of the logical table
        sub_type: string literal used to identify the type of a metadata object,
            ObjectSubType enum
        version: object version ID
        ancestors: list of ancestor folders
        type: object type, ObjectTypes enum
        ext_type: object extended type, ExtendedType enum
        date_created: creation time, DateTime object
        date_modified: last modification time, DateTime object
        owner: User object that is the owner
        acg: access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: object access control list
        version_id: the version number this object is currently carrying
        is_embedded: if true indicates that the target object of this reference
            is embedded within this object, if this field is omitted
            (as is usual) then the target is not embedded.
        path: the path of the object, read only
        primary_locale: the primary locale of the object, in the IETF BCP 47
            language tag format, such as "en-US", read only
        destination_folder_id: a globally unique identifier used to distinguish
            between metadata objects within the same project
        logical_size: a size based on the columns in the tables and the
            attributes to which those columns correspond. Uses the conceptual or
            logical attribute definitions to assign a size to each table in the
            project.
        is_logical_size_locked: An indication if the logical table size of a
            table is locked or not. When a table’s logical size is locked, the
            table is excluded from the logical table size calculation when a
            schema update is performed.
        is_part_of_partition: An indication whether this table is part of
            a partition table.
        is_true_key: An indication whether a specified primary key is the true
            key.
        table_key: A list of objects representing table keys defined on physical
            table
        partition_level: List of objects representing partition attributes
        attributes: list of objects representing business concepts reflected in
            data. Attributes provide a context in which to report on and analyze
            business facts or calculations.
        facts: list of objects representing essential elements within the
            business data model and basis for almost all metrics. They relate
            numeric data values from the data warehouse to the MicroStrategy
            reporting environment. E.g. sales dollars, units sold, profit etc.
        primary_data_source: Table's data can be from different data sources.
            Primary data source is the first option for table's data source.
        enclose_sql_in_parentheses: This property is only applied if a logical
            table is based upon a free form sql table. By default, this property
            value is set to true. This encloses the entire SQL that you typed in
            the sqlStatement field in parentheses when the SQL statement is
            executed.
        physical_table: Metadata of a physical table based on warehouse's
            physical table or free form sql.
        partition_tables: A list of objects representing partition's tables
            metadata
        secondary_data_sources: MicroStrategy support mapping the table to more
            than one data source. This attribute is a list of object
            representations of a reference to a datasource.
        hidden: Specifies whether the object is hidden
    """

    _OBJECT_TYPE: ObjectTypes = ObjectTypes.TABLE
    _FROM_DICT_MAP: dict = {
        **Entity._FROM_DICT_MAP,
        "owner": User.from_dict,
        "attributes": [Attribute.from_dict],
        "sub_type": ObjectSubType,
        "primary_data_source": SchemaObjectReference.from_dict,
        "secondary_data_sources": [SchemaObjectReference.from_dict],
        "physical_table": PhysicalTable.from_dict,
        "table_key": [SchemaObjectReference.from_dict],
        "facts": [Fact.from_dict],
    }
    _API_GETTERS = {
        (
            "name",
            "description",
            "abbreviation",
            "type",
            "subtype",
            "ext_type",
            "date_created",
            "date_modified",
            "version",
            "owner",
            "icon_path",
            "view_media",
            "ancestors",
            "certified_info",
            "acg",
            "acl",
            "comments",
            "project_id",
            "hidden",
            "target_info",
            'hidden',
            'comments',
        ): objects_processors.get_info,
        (
            "id",
            "is_embedded",
            "primary_locale",
            "destination_folder_id",
            "logical_size",
            "is_logical_size_locked",
            "is_part_of_partition",
            "is_true_key",
            "table_key",
            "partition_level",
            "attributes",
            "facts",
            "primary_data_source",
            "enclose_sql_in_parentheses",
            "physical_table",
            "partition_tables",
            "secondary_data_sources",
        ): tables_api.get_table,
    }
    _API_PATCH: dict = {
        (
            "name",
            "description",
            "logical_size",
            "is_logical_size_locked",
            "is_true_key",
            "primary_data_source",
            "enclose_sql_in_parentheses",
            "physical_table",
            "secondary_data_sources",
        ): (tables_api.patch_table, "partial_put"),
        (
            "folder_id",
            "hidden",
            "comments",
        ): (objects_processors.update, "partial_put"),
    }
    _PATCH_PATH_TYPES = {
        "name": str,
        "description": str,
        "logical_size": int,
        "is_logical_size_locked": bool,
        "is_true_key": bool,
        "primary_data_source": dict,
        "enclose_sql_in_parentheses": bool,
        "physical_table": dict,
        "secondary_data_sources": list,
        "folder_id": str,
        "hidden": bool,
    }

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ):
        """Initializes a LogicalTable object. You can specify either `id`
            or `name` where `id` has a priority. You must specify
            either `project_id` or `project_name` if Connection object provided
            as `connection` is not mapped to a project.

        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted.

        Note:
            When `project_id` is `None` and `project_name` is `None`, then
            its value is overwritten by `project_id` from `connection` object.

        Args:
            connection (Connection): Object representation of MSTR Connection.
            id (Optional[str], optional): ID of a table. Defaults to None.
            name (Optional[str], optional): Name of a table. Defaults to
                None.
            project_id (Optional[str], optional): ID of a project. Defaults to
                None.
            project_name (Optional[str], optional): Name of a project. Defaults
                to None.

        Throws:
            ValueError if:
                - neither `id` nor `name` were specified
                - a specified project could not be found.
                - a specified table could not be found.

        Examples:
            >>> LogicalTable(connection,
                             id='FF529E1446BE3DF75ECD4B8E18618747')
            >>> LogicalTable(connection, name='lu_day_of_week')
        """

        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=not project_name,
        )

        if not id and not name:
            exception_handler(
                msg="You must specify table's id or table's name.",
                exception_type=ValueError,
            )
        elif not id:
            try:
                id = full_search(
                    connection=connection,
                    project=project_id,
                    object_types=ObjectTypes.TABLE,
                    name=name,
                )[0].get("id")
            except IndexError:
                exception_handler(
                    msg="Table with a given name was not found.",
                    exception_type=ValueError,
                )
        super().__init__(connection=connection, object_id=id)

    def _init_variables(self, default_value, **kwargs) -> None:
        super()._init_variables(default_value=default_value, **kwargs)

        attributes = kwargs.get('attributes')
        self.attributes = (
            Attribute.bulk_from_dict(source_list=attributes, connection=self.connection)
            if attributes
            else default_value
        )

        ancestors = kwargs.get('ancestors')
        self.folder_id = ancestors[-1].get('id') if ancestors else default_value
        self.destination_folder_id = kwargs.get('destination_folder_id', default_value)

        facts = kwargs.get('facts')
        if facts:
            [f.update(f.pop('information')) for f in facts]
        self.facts = (
            Fact.bulk_from_dict(source_list=facts, connection=self.connection)
            if facts
            else default_value
        )
        self.is_logical_size_locked = kwargs.get(
            'is_logical_size_locked', default_value
        )
        self.is_part_of_partition = kwargs.get('is_part_of_partition', default_value)
        self.is_true_key = kwargs.get('is_true_key', default_value)
        self.logical_size = kwargs.get('logical_size', default_value)

        physical_table = kwargs.get('physical_table')
        self.physical_table = (
            PhysicalTable.from_dict(source=physical_table, connection=self.connection)
            if physical_table
            else default_value
        )

        primary_data_source = kwargs.get('primary_data_source')
        self.primary_data_source = (
            SchemaObjectReference.from_dict(source=primary_data_source)
            if primary_data_source
            else default_value
        )
        self.primary_locale = kwargs.get('primary_locale', default_value)

        secondary_data_sources = kwargs.get('secondary_data_sources')
        self.secondary_data_sources = (
            SchemaObjectReference.bulk_from_dict(
                source_list=secondary_data_sources, connection=self.connection
            )
            if secondary_data_sources
            else default_value
        )

        sub_type = kwargs.get('sub_type')
        self._sub_type = ObjectSubType(sub_type) if sub_type else default_value

        table_key = kwargs.get('table_key')
        self.table_key = (
            SchemaObjectReference.bulk_from_dict(
                source_list=table_key, connection=self.connection
            )
            if table_key
            else default_value
        )
        self._version = kwargs.get('version_id')

    @classmethod
    def create(
        cls,
        connection: "Connection",
        primary_data_source: SchemaObjectReference | dict,
        table_name: str | None = None,
        table_description: str | None = None,
        sub_type: ObjectSubType | str | None = ObjectSubType.LOGICAL_TABLE,
        destination_folder: Folder | str | None = None,
        is_embedded: bool | None = None,
        physical_table: PhysicalTable | dict | None = None,
        physical_table_name: str | None = None,
        physical_table_namespace: str | None = None,
        physical_table_type: PhysicalTableType | None = PhysicalTableType.NORMAL,
        physical_table_prefix: str | None = None,
        columns: list[TableColumn] | list[dict] | None = None,
        sql_statement: str | None = None,
        logical_size: int | None = None,
        is_part_of_partition: bool | None = None,
        is_true_key: bool | None = None,
        enclose_sql_in_parentheses: bool | None = None,
        check_secondary_data_source_table: bool | None = None,
        column_merge_option: None | (
            TableColumnMergeOption
        ) = TableColumnMergeOption.REUSE_ANY,
        table_prefix_option: TablePrefixOption | None = None,
    ) -> "LogicalTable":
        """Create a new table in a specific project.

        Args:
            connection (object): MicroStrategy connection object returned by
                `connection.Connection()`.
            primary_data_source (object): Information about an object referenced
                within the  specification of another object. An object reference
                typically contains only enough fields to uniquely identify
                the referenced objects.
            table_name (str, optional): Logical table name
            table_description (str, optional): Logical table description
            sub_type (enum, str, optional): String literal used to identify the
                type of a metadata object.
            destination_folder (object, str, optional): Globally unique
                identifier used to distinguish between metadata objects within
                the same project
            is_embedded (bool, optional): if true indicates that the target
                object of this reference is embedded within this object
            physical_table (object, dict, optional): Physical table
            physical_table_name (str, optional): Table name
            physical_table_namespace (str, optional): Datasource namespace
            physical_table_type (enum, optional): Table type
            physical_table_prefix (str, optional): Table prefix of the table
            columns (object, optional): Columns
            sql_statement (str, optional): SQL statement that will be executed
                against the warehouse to get column data of the physical table.
            logical_size (int, optional): Size of a table
            is_part_of_partition (bool, optional): whether current table is part
                of partition table.
            is_true_key (bool, optional): MicroStrategy requires each table to
                have a primary key, which is a unique value identifying each
                distinct data record or row. A primary key can be defined by one
                or more columns in the table. MicroStrategy determines the
                primary key for a table based on the attribute's mapped to the
                columns of the table. The key is made up of the lowest level
                attributes. If these columns are mapped to attributes in
                MicroStrategy, then the primary key is represented correctly.
                In this case, if isTrueKey is returned as true, it indicates
                specified primary key is the true key.
            enclose_sql_in_parentheses (bool, optional): This property is only
                applied to free form sql table. This encloses the entire SQL
                that you typed in the sqlStatement field in parentheses when
                the SQL statement is executed.
            check_secondary_data_source_table (bool, optional): API finds
                compatible tables in the project. If a compatible table is
                found, the compatible table object information is returned.
                If no table is found, a new table is created.
                Available values: 'true', 'false'
                    If 'true', finds compatible tables in the project.
                        If a compatible table is found, the compatible table
                        object information is returned. If no table is found,
                        a new table is created.
                    If 'false', a new table is created.
            column_merge_option (enum, optional): Defines a column merge option
                Available values: 'reuse_any', 'reuse_compatible_data_type',
                'reuse_matched_data_type'.
                If 'TableColumnMergeOption.REUSE_ANY', updates the column
                    data type to use the most recent column definition.
                If 'TableColumnMergeOption.REUSE_COMPATIBLE_DATA_TYPE', updates
                    the column data type to use the data type with the largest
                    precision or scale.
                If 'TableColumnMergeOption.REUSE_MATCHED_DATA_TYPE', renames
                    the column in newly added table to allow it to have
                    different data types.
            table_prefix_option (enum, optional): Define the table prefix
                Available values: 'ADD_DEFAULT_PREFIX', 'ADD_NAMESPACE'.
                If 'TablePrefixOption.ADD_DEFAULT_PREFIX', applies the default
                    prefix setting on warehouse catalog.
                If 'TablePrefixOption.ADD_NAMESPACE', create a prefix same
                    with namespace.

        Returns:
            LogicalTable object

        Examples:
            >>> LogicalTable.create(connection,
                                    physical_table=PhysicalTable(
                                        connection=connection,
                                        id="8D67911E11D3E4981000E787EC6DE8A4",
                                    ),
                                    primary_data_source=SchemaObjectReference(
                                        object_id="A23BBC514D336D5B4FCE919FA3",
                                        sub_type=ObjectSubType.DB_ROLE
                                    ),
                                    )
            >>> LogicalTable.create(connection,
                    physical_table_name="lu_prescriber",
                    physical_table_namespace="public",
                    primary_data_source=SchemaObjectReference(
                            object_id="DAD6CAD6457DAF29E34463961688EA60",
                            sub_type=ObjectSubType.DB_ROLE),
                    table_name="New Table", table_description="New Description",
                    sub_type=ObjectSubType.LOGICAL_TABLE))
        """
        if physical_table and any(
            [
                physical_table_name,
                physical_table_namespace,
                physical_table_prefix,
                sql_statement,
                columns,
            ]
        ):
            msg = (
                "You can either pass the `physical table` object or specify "
                "parameters related to the physical table, not both."
            )
            exception_handler(msg, exception_type=ValueError)

        physical_table = (
            PhysicalTable(connection=connection, id=physical_table.get("id"))
            if isinstance(physical_table, dict)
            else physical_table
        )

        if get_enum_val(
            getattr(physical_table, "table_type", physical_table_type)
        ) == get_enum_val(PhysicalTableType.WAREHOUSE_PARTITION):
            exception_handler(
                msg="Warehouse Partition tables are not supported.",
                exception_type=TypeError,
            )

        if (physical_table_name and physical_table_namespace) and sql_statement:
            msg = (
                "Specify either a `physical_table_name` and `physical_table_namespace`"
                " or `sql_statement`"
            )
            exception_handler(msg, exception_type=ValueError)

        primary_data_source = (
            primary_data_source.to_dict()
            if isinstance(primary_data_source, SchemaObjectReference)
            else primary_data_source
        )

        physical_table = {
            "type": get_enum_val(
                getattr(physical_table, "table_type", physical_table_type),
                PhysicalTableType,
            ),
            "tableName": getattr(physical_table, "table_name", physical_table_name),
            "namespace": getattr(physical_table, "namespace", physical_table_namespace),
            "tablePrefix": getattr(
                physical_table, "table_prefix", physical_table_prefix
            ),
            "columns": cls.__prepare_columns(
                getattr(physical_table, "columns", columns)
            ),
            "sqlStatement": getattr(physical_table, "sql_statement", sql_statement),
        }

        body = {
            "information": {
                "name": table_name,
                "description": table_description,
                "subType": get_enum_val(sub_type, ObjectSubType),
                "destinationFolderId": (
                    destination_folder.id
                    if isinstance(destination_folder, Folder)
                    else destination_folder
                ),
                "isEmbedded": is_embedded,
            },
            "logicalSize": logical_size,
            "isPartOfPartition": is_part_of_partition,
            "isTrueKey": is_true_key,
            "primaryDataSource": primary_data_source,
            "encloseSqlInParentheses": enclose_sql_in_parentheses,
            "physicalTable": physical_table,
        }
        body = delete_none_values(body, recursion=True)

        response = tables_api.post_table(
            connection=connection,
            data=body,
            check_secondary_data_source_table=check_secondary_data_source_table,
            column_merge_option=get_enum_val(
                column_merge_option, TableColumnMergeOption
            ),
            table_prefix_option=get_enum_val(table_prefix_option, TablePrefixOption),
        ).json()

        if config.verbose:
            logger.info(
                f"Successfully created table named: '{response['name']}' "
                f"with ID: '{response['id']}'"
            )

        table = cls.from_dict(source=response, connection=connection)

        # destination_folder attribute is ignored by Modeling Service
        # logical table object is always placed in the default folder
        # so move to different folder after creation
        if destination_folder:
            table.move(destination_folder)

        return table

    @classmethod
    def create_alias(cls, connection: "Connection", id: str) -> type["LogicalTable"]:
        """Create a new table alias in a specific project.

        Args:
            connection (object): MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Physical table id
        Returns:
            LogicalTable object

        Examples:
            >>> LogicalTable.create_alias(connection,
                                          id="E278D17342991E49710D6F90E2A7BF2C"
                                          )
        """
        body = {"physicalTable": {"information": {"objectId": id}}}

        response = tables_api.post_table(connection=connection, data=body).json()
        if config.verbose:
            logger.info(
                f"Successfully created table alias named: '{response['name']}' "
                f"with ID: '{response['id']}'"
            )

        return cls.from_dict(source=response, connection=connection)

    def alter(
        self,
        name: str | None = None,
        is_true_key: bool | None = None,
        logical_size: int | None = None,
        description: str | None = None,
        is_logical_size_locked: bool | None = None,
        primary_data_source: SchemaObjectReference | None = None,
        secondary_data_sources: None | (
            list[SchemaObjectReference] | list[dict]
        ) = None,
        physical_table_object_name: str | None = None,
        physical_table_name: str | None = None,
        physical_table_prefix: str | None = None,
        sql_statement: str | None = None,
        enclose_sql_in_parentheses: bool | None = None,
        columns: list[TableColumn] | list[dict] | None = None,
        folder_id: str | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
    ) -> None:  # NOSONAR
        """Alters properties specified by keyword arguments.

        Args:
            name (Optional[str], optional): Name of a logical table. Defaults to
                None.
            is_true_key (Optional[bool]): MicroStrategy requires each table to
                have a primary key, which is a unique value identifying each
                distinct data record or row. A primary key can be defined by one
                or more columns in the table. MicroStrategy determines the
                primary key for a table based on the attribute's mapped to the
                columns of the table. The key is made up of the lowest level
                attributes. If these columns are mapped to attributes in
                MicroStrategy, then the primary key is represented correctly.
                In this case, if isTrueKey is returned as true, it indicates
                specified primary key is the true key. Defaults to None.
            logical_size (Optional[int], optional): Size of a logical table.
                Defaults to None.
            description (str, optional): Logical table description
            is_logical_size_locked (Optional[bool], optional): An indication if
                the logical table size of a table is locked or not. When a
                table’s logical size is locked, the table is excluded from the
                logical table size calculation when a schema update is
                performed. Defaults to None.
            primary_data_source (Optional[SchemaObjectReference], optional):
                Table's data can be from different data sources. Primary data
                source is the first option for table's data source. Defaults to
                None.
            secondary_data_sources (Optional[List[SchemaObjectReference] |
                List[dict]], optional): MicroStrategy support mapping the table
                to more than one data source. This attribute is a list of object
                representations of a reference to a datasource. Defaults to None
            physical_table_object_name (Optional[str], optional): A new name of
                an object (on MSTR server) that represents a physical table
                mapped to the logical table. Defaults to None.
            physical_table_name (Optional[str], optional): A new name of a
                physical table mapped to the logical table. Defaults to None.
            physical_table_prefix (Optional[str], optional): A new prefix for a
                physical table mapped to the logical table. Defaults to None.
            sql_statement (Optional[str], optional): A new SQL statement for
                physical table mapped to the logical table. Defaults to None.
            enclose_sql_in_parentheses(Optional[bool], optional): This encloses
                the entire SQL that you typed in the sqlStatement field
                in parentheses when the SQL statement is executed.
            columns (Optional[List[TableColumn] | List[dict]], optional):
                A list of new columns of a physical table mapped to the logical
                table. Defaults to None.
            folder_id (Optional[str], optional): The ID of a folder to which a
                logical  table should be moved. Defaults to None.
            hidden (bool, optional): Specifies whether the object is hidden.
                Default value: False.
            comments (str, optional): Comments added to the object. Defaults to
                None.

            Throws:
                TypeError if:
                    - Attempting to alter a logical table when it depends upon a
                        physical table which type is "Warehouse Partition Table"
                    - Attempting to alter `physical_table_name` or
                        `physical_table_prefix` when the logical table depends
                        upon a freeform sql physical table.
                    - Attempting to alter `sql_statement` or columns when the
                        logical table depends upon a normal physical table.

            Examples:
                >>> LogicalTable(connection,
                                 id='FF529E1446BE3DF75ECD4B8E18618747').alter(
                                     name="Altered Name"
                                 )
        """
        self.__validate_physical_table_type()
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]

        physical_table_type = self.physical_table.table_type
        if physical_table_type == PhysicalTableType.NORMAL and (
            sql_statement or columns
        ):
            exception_handler(
                msg=(
                    "You cannot change `sql_statement` nor `columns` attributes when "
                    "a logical table depends upon a normal physical table."
                ),
                exception_type=TypeError,
            )
        elif physical_table_type == PhysicalTableType.SQL and (
            physical_table_prefix or physical_table_name
        ):
            exception_handler(
                msg=(
                    "You cannot change `physical_table_name` nor "
                    "`physical_table_prefix` when a logical table depends upon "
                    "a freeform sql physical table."
                ),
                exception_type=TypeError,
            )

        physical_table = (
            {"information": {"name": properties.pop("physical_table_object_name")}}
            if properties.get("physical_table_object_name")
            else {}
        )
        physical_table.update(
            {
                "table_prefix": properties.pop("physical_table_prefix", None),
                "table_name": properties.pop("physical_table_name", None),
                "sql_statement": properties.pop("sql_statement", None),
                "columns": self.__prepare_columns(properties.pop("columns", None)),
            }
        )
        properties["physical_table"] = delete_none_values(
            physical_table, recursion=True
        )

        self._alter_properties(**properties)

    def update_physical_table_structure(
        self,
        col_merge_option: TableColumnMergeOption | str | None = None,
        ignore_table_prefix: bool | None = None,
    ) -> None:
        """Updates a structure of a physical table upon which the logical table
           depends.

        Args:
            col_merge_option (enum, optional): Defines a column merge option
                Available values: 'reuse_any', 'reuse_compatible_data_type',
                'reuse_matched_data_type'.
                If 'TableColumnMergeOption.REUSE_ANY', updates the column
                    data type to use the most recent column definition.
                If 'TableColumnMergeOption.REUSE_COMPATIBLE_DATA_TYPE',
                    updates the column data type to use the data type with
                    the largest precision or scale.
                If 'TableColumnMergeOption.REUSE_MATCHED_DATA_TYPE', renames
                    the column in newly added table to allow it to have
                    different data types.
                If this value is not set, use the option that set on DBRole
                catalog setting.

            ignore_table_prefix (bool, optional): If true, get all tables under
                current DB. There are three following situations:

                    - If there is only one table that has same name as updated
                        table, update table structure using this table.
                    - If there is no table that has same name as updated table,
                        throw error.
                    - If there are multiple tables has same name as updated
                        table, throw error.

                If false, remain the current behavior.

                If not set, get the setting value from warehouse catalog. This
                behavior is same as column merge options.

        Examples:
            >>> LogicalTable(
            >>>     connection, id='FF529E1446BE3DF75ECD4B8E18618747'
            >>> ).update_physical_table_structure(
            >>>     col_merge_option=TableColumnMergeOption.REUSE_ANY
            >>> )
        """
        col_merge_option = (
            get_enum_val(col_merge_option, TableColumnMergeOption)
            if col_merge_option
            else None
        )

        table_data = tables.update_structure(
            self.connection, self.id, col_merge_option, ignore_table_prefix
        )

        self._set_object_attributes(**table_data)

        data = table_data.get("physicalTable")

        logger.info(
            "Successfully modified a structure of a physical table '"
            "%s' with ID: '%s' in a '%s' namespace with a '%s' table prefix.",
            data.get('tableName'),
            data.get('information').get('objectId'),
            data.get('namespace'),
            data.get('tablePrefix'),
        )

    @classmethod
    def update_physical_table_structure_for_all_tables(
        cls,
        connection: "Connection",
        col_merge_option: TableColumnMergeOption | str | None = None,
        ignore_table_prefix: bool | None = None,
    ):
        """Updates structure for every table in a project mapped to a
           connection.

        Args:
            connection (Connection): Object representation of MSTR Connection.
            col_merge_option (enum, optional): Defines a column merge option
                Available values: 'reuse_any', 'reuse_compatible_data_type',
                'reuse_matched_data_type'.
                If 'TableColumnMergeOption.REUSE_ANY', updates the column
                    data type to use the most recent column definition.
                If 'TableColumnMergeOption.REUSE_COMPATIBLE_DATA_TYPE',
                    updates the column data type to use the data type with
                    the largest precision or scale.
                If 'TableColumnMergeOption.REUSE_MATCHED_DATA_TYPE', renames
                    the column in newly added table to allow it to have
                    different data types.
                If this value is not set, use the option that set on DBRole
                catalog setting.

            ignore_table_prefix (bool, optional): If true, get all tables under
                current DB. There are three following situations:

                    - If there is only one table that has same name as updated
                        table, update table structure using this table.
                    - If there is no table that has same name as updated table,
                        throw error.
                    - If there are multiple tables has same name as updated
                        table, throw error.

                If false, remain the current behavior.

                If not set, get the setting value from warehouse catalog. This
                behavior is same as column merge options.
        """
        col_merge_option = get_enum(col_merge_option, TableColumnMergeOption)
        logical_tables = list_logical_tables(connection)
        with tqdm(
            total=len(logical_tables), desc="Updating structures..."
        ) as progress_bar:
            for table in logical_tables:
                table.update_physical_table_structure(
                    col_merge_option=col_merge_option,
                    ignore_table_prefix=ignore_table_prefix,
                )
                progress_bar.update()

    def __validate_physical_table_type(self) -> bool:
        """Validates whether a table can be modified by checking if a physical
            table type upon which the logical table depends is
            "Warehouse Partition Table"

        Returns:
            bool: True if the type of physical table is not
                "Warehouse Partition Table", else False
        """
        if self.physical_table.table_type == PhysicalTableType.WAREHOUSE_PARTITION:
            exception_handler(
                msg="Warehouse Partition tables cannot be edited.",
                exception_type=TypeError,
            )
        return True

    @staticmethod
    def __prepare_columns(
        columns: list[TableColumn] | list[dict] | None,
    ) -> list[dict] | None:
        """Prepares columns for sending requests

        Args:
            columns (list[TableColumn] | list[dict] | None): Columns to prepare

        Returns:
            list[dict]: Prepared columns
        """
        if not columns:
            return None

        columns = [
            col.to_dict() if isinstance(col, TableColumn) else col for col in columns
        ]
        return [
            {
                "information": {
                    "objectId": column.get("id"),
                    "name": column.get("name"),
                    "subType": column.get("subType"),
                    "dateCreated": column.get("dateCreated"),
                    "dateModified": column.get("dateModified"),
                    "versionId": column.get("versionId"),
                    "primaryLocale": column.get("primaryLocale"),
                },
                "columnName": column.get("columnName"),
                "dataType": column.get("dataType"),
            }
            for column in columns
        ]

    def list_columns(self, to_dictionary: bool = False):
        """Lists columns of the table in a dictionary or objects form

        Args:
        to_dictionary (bool): specifies whether to list the columns as
            dictionaries or objects. Defaults to False

        Returns:
            A list of columns as dictionaries or objects, depending on input"""
        if to_dictionary:
            return [column.to_dict() for column in self.physical_table.columns]
        return self.physical_table.columns
